from ylz_utils.langchain.graph import GraphLib

from langgraph.graph import StateGraph, START,END, MessagesState
from langchain_core.messages import AIMessage
from langgraph.graph.state import CompiledStateGraph

from langchain_community.utilities import SQLDatabase
from typing import Any
from langchain_core.messages import ToolMessage
from langchain_core.runnables import RunnableLambda, RunnableWithFallbacks
from langgraph.prebuilt import ToolNode

from langchain_core.tools import tool

from langchain_community.agent_toolkits import SQLDatabaseToolkit

from langchain_core.prompts import ChatPromptTemplate


from typing import Annotated, Literal

from pydantic import BaseModel, Field
from typing_extensions import TypedDict

from langgraph.graph.message import AnyMessage, add_messages

class State(TypedDict):
    messages: Annotated[list[AnyMessage], add_messages]
class SubmitFinalAnswer(BaseModel):
    """Submit the final answer to the user based on the query results."""
    final_answer: str = Field(..., description="The final answer to the user")

class DbGraph(GraphLib):
    db = None
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
    def set_db(self,uri:str):
        self.db = SQLDatabase.from_uri(uri)
    def _check(self):
        if not self.db:
            raise Exception("先调用set_db(uri)设置db")
        if not self.llm:
            raise Exception("先调用set_llm(key,model)设置llm")
    def set_toolkit(self):
        self._check()
        toolkit = SQLDatabaseToolkit(db=self.db, llm=self.llm)
        tools = toolkit.get_tools()
        self.list_tables_tool = next(tool for tool in tools if tool.name == "sql_db_list_tables")
        self.get_schema_tool = next(tool for tool in tools if tool.name == "sql_db_schema")
    @tool
    def db_query_tool(self,query: str) -> str:
        """
        Execute a SQL query against the database and get back the result.
        If the query is not correct, an error message will be returned.
        If an error is returned, rewrite the query, check the query, and try again.
        """
        self._check()
        result = self.db.run_no_throw(query)
        if not result:
            return "Error: Query failed. Please rewrite your query and try again."
        return result
   
    def create_tool_node_with_fallback(self,tools: list) -> RunnableWithFallbacks[Any, dict]:
        """
        Create a ToolNode with a fallback to handle errors and surface them to the agent.
        """
        return ToolNode(tools).with_fallbacks(
            [RunnableLambda(self.handle_tool_error)], exception_key="error"
        )
    def handle_tool_error(self,state) -> dict:
        error = state.get("error")
        tool_calls = state["messages"][-1].tool_calls
        return {
            "messages": [
                ToolMessage(
                    content=f"Error: {repr(error)}\n please fix your mistakes.",
                    tool_call_id=tc["id"],
                )
                for tc in tool_calls
            ]
        }
    def first_tool_call(self,state: State) -> dict[str, list[AIMessage]]:
        return {
            "messages": [
                AIMessage(
                    content="",
                    tool_calls=[
                        {
                            "name": "sql_db_list_tables",
                            "args": {},
                            "id": "tool_abcd123",
                        }
                    ],
                )
            ]
        }
    def model_check_query(self,state: State) -> dict[str, list[AIMessage]]:
        """
        Use this tool to double-check if your query is correct before executing it.
        """
        self._check()
        query_check_system = """You are a SQL expert with a strong attention to detail.
        Double check the SQLite query for common mistakes, including:
        - Using NOT IN with NULL values
        - Using UNION when UNION ALL should have been used
        - Using BETWEEN for exclusive ranges
        - Data type mismatch in predicates
        - Properly quoting identifiers
        - Using the correct number of arguments for functions
        - Casting to the correct data type
        - Using the proper columns for joins

        If there are any of the above mistakes, rewrite the query. If there are no mistakes, just reproduce the original query.

        You will call the appropriate tool to execute the query after running this check."""

        query_check_prompt = ChatPromptTemplate.from_messages(
            [("system", query_check_system), ("placeholder", "{messages}")]
        )
        query_check = query_check_prompt | self.llm.bind_tools(
            [self.db_query_tool], tool_choice="required"
        )
        return {"messages": [query_check.invoke({"messages": [state["messages"][-1]]})]}

    def query_gen_node(self,state: State):
        self._check()
        # Add a node for a model to generate a query based on the question and schema
        query_gen_system = """You are a SQL expert with a strong attention to detail.

        Given an input question, output a syntactically correct SQLite query to run, then look at the results of the query and return the answer.

        DO NOT call any tool besides SubmitFinalAnswer to submit the final answer.

        When generating the query:

        Output the SQL query that answers the input question without a tool call.

        Unless the user specifies a specific number of examples they wish to obtain, always limit your query to at most 5 results.
        You can order the results by a relevant column to return the most interesting examples in the database.
        Never query for all the columns from a specific table, only ask for the relevant columns given the question.

        If you get an error while executing a query, rewrite the query and try again.

        If you get an empty result set, you should try to rewrite the query to get a non-empty result set. 
        NEVER make stuff up if you don't have enough information to answer the query... just say you don't have enough information.

        If you have enough information to answer the input question, simply invoke the appropriate tool to submit the final answer to the user.

        DO NOT make any DML statements (INSERT, UPDATE, DELETE, DROP etc.) to the database."""

        query_gen_prompt = ChatPromptTemplate.from_messages(
            [("system", query_gen_system), ("placeholder", "{messages}")]
        )
        query_gen = query_gen_prompt | self.llm.bind_tools(
            [SubmitFinalAnswer]
        )
        message = query_gen.invoke(state)

        # Sometimes, the LLM will hallucinate and call the wrong tool. We need to catch this and return an error message.
        tool_messages = []
        if message.tool_calls:
            for tc in message.tool_calls:
                if tc["name"] != "SubmitFinalAnswer":
                    tool_messages.append(
                        ToolMessage(
                            content=f"Error: The wrong tool was called: {tc['name']}. Please fix your mistakes. Remember to only call SubmitFinalAnswer to submit the final answer. Generated queries should be outputted WITHOUT a tool call.",
                            tool_call_id=tc["id"],
                        )
                    )
        else:
            tool_messages = []
        return {"messages": [message] + tool_messages}
    def should_continue(self,state: State) -> Literal[END, "correct_query", "query_gen"]:
        messages = state["messages"]
        last_message = messages[-1]
        # If there is a tool call, then we finish
        if getattr(last_message, "tool_calls", None):
            return END
        if last_message.content.startswith("Error:"):
            return "query_gen"
        else:
            return "correct_query"
        
    def get_graph(self) -> CompiledStateGraph:
        # Define a new graph
        self._check()
        self.set_toolkit()
        
        workflow = StateGraph(State)
        workflow.add_node("first_tool_call", self.first_tool_call)
        workflow.add_node(
            "list_tables_tool", self.create_tool_node_with_fallback([self.list_tables_tool])
        )
        workflow.add_node("get_schema_tool", self.create_tool_node_with_fallback([self.get_schema_tool]))
        model_get_schema = self.llm.bind_tools(
            [self.get_schema_tool]
        )
        workflow.add_node(
            "model_get_schema",
            lambda state: {
                "messages": [model_get_schema.invoke(state["messages"])],
            },
        )
        workflow.add_node("query_gen", self.query_gen_node)
        workflow.add_node("correct_query", self.model_check_query)
        workflow.add_node("execute_query", self.create_tool_node_with_fallback([self.db_query_tool]))

        workflow.add_edge(START, "first_tool_call")
        workflow.add_edge("first_tool_call", "list_tables_tool")
        workflow.add_edge("list_tables_tool", "model_get_schema")
        workflow.add_edge("model_get_schema", "get_schema_tool")
        workflow.add_edge("get_schema_tool", "query_gen")
        workflow.add_conditional_edges(
            "query_gen",
            self.should_continue,
        )
        workflow.add_edge("correct_query", "execute_query")
        workflow.add_edge("execute_query", "query_gen")

        graph = workflow.compile(checkpointer=self.memory)
        return graph 
    


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from abc import ABC,abstractmethod

from operator import itemgetter
from typing import Literal,List,Annotated
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage
from pydantic import BaseModel
from typing_extensions import TypedDict
from langchain_core.tools import tool,Tool
from langchain_core.runnables import RunnablePassthrough,RunnableLambda,RunnableParallel
from langgraph.graph import START,END,StateGraph,MessagesState
from langgraph.checkpoint.memory import MemorySaver
#from langgraph.checkpoint.sqlite import SqliteSaver
#from langgraph.checkpoint.aiosqlite import AsyncSqliteSaver

from langgraph.prebuilt import ToolExecutor,ToolInvocation, tools_condition
from langgraph.graph.state import CompiledStateGraph

from ylz_utils.file import FileLib
from ylz_utils.data import StringLib,Color

class GraphLib(ABC):
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.user_id = 'default'
        self.conversation_id = 'default'
        self.thread_id='default-default'
        self.nodes_llm_config = {}
        self.websearch_tool = None
        self.ragsearch_tool = None
        self.python_repl_tool = langchainLib.get_python_repl_tool()
        self.chat_db_name = ":memory:"
        self.memory = MemorySaver()
        #self.memory = SqliteSaver.from_conn_string(self.chat_db_name)
        #self.amemory = AsyncSqliteSaver.from_conn_string(self.chat_db_name)
        self.query_dbname = None
        self.tools=[]
        self.tools_executor = None

    def set_chat_dbname(self,dbname):
        # "checkpoint.sqlite"
        self.chat_dbname = dbname
        self.memory = SqliteSaver.from_conn_string(dbname)
    def set_query_dbname(self,dbname):
        self.query_dbname = dbname
    def set_ragsearch_tool(self,retriever):
        name = "rag_searcher"
        description = "一个有用的工具用来从本地知识库中获取信息。你总是利用这个工具优先从本地知识库中搜索有用的信息"
        self.ragsearch_tool = self.langchainLib.get_ragsearch_tool(retriever,name,description)
    def set_websearch_tool(self,websearch_key):
        self.websearch_tool = self.langchainLib.get_websearch_tool(websearch_key)
    def set_tools(self,tools=None):
        if not tools:
            tools = [
                self.python_repl_tool,
            ]
            if self.websearch_tool:
                tools.append(self.websearch_tool)
            if self.ragsearch_tool:
                tools.append(self.ragsearch_tool)
        self.tools = tools
        return self.tools    
    def set_tools_executor(self,tools):
        if not isinstance(tools,(list,tuple)):
            tools = [tools]
        self.tools_executor = ToolExecutor(tools) 
        return self.tools_executor
       
    def create_response(self, response: str, ai_message: AIMessage):
        return ToolMessage(
            content = response,
            tool_call_id = ai_message.tools_calls[0]["id"]
        )

    def set_nodes_llm_config(self,nodes_llm_config:dict[str,dict[Literal["llm_key","llm_model"],str|None]]|tuple):
        # {
        #     "node1":{"llm_key":...,"llm_mode":...}
        #     "node2":{"llm_key":...,"llm_mode":...}
        # }
        if isinstance(nodes_llm_config,tuple):
            self.nodes_llm_config['default'] = {"llm_key":nodes_llm_config[0],"llm_model":nodes_llm_config[1]}
        else:
            self.nodes_llm_config = nodes_llm_config
    def get_node_llm(self,node_key=None):
        try:
            node_llm_config:dict = self.nodes_llm_config[node_key] 
            llm_key = node_llm_config.get("llm_key")
            llm_model = node_llm_config.get("llm_model")
            llm = self.langchainLib.get_llm(key=llm_key,model=llm_model)
        except:            
            default_llm_config = self.nodes_llm_config.get('default',{})
            if default_llm_config:
                llm = self.langchainLib.get_llm(key = default_llm_config.get('llm_key'), model = default_llm_config.get('llm_model'))
            else:
                llm = self.langchainLib.get_llm()
        finally:
            #print("llm=",llm.model_name,llm.openai_api_base)
            return llm
    def set_thread(self,user_id="default",conversation_id="default"):
        self.user_id = user_id
        self.conversation_id = conversation_id
        self.thread_id = f"{user_id}-{conversation_id}"

    def tool_node(self,state):
        if not self.tools_executor:
            raise Exception("please call set_tools_executor(tools) first!")
        messages = state["messages"]
        last_message = messages[-1]
        tool_call = last_message.tool_calls[0]
        action = ToolInvocation(
            tool = tool_call["name"],
            tool_input = tool_call["args"]
        )  
        response = self.tools_executor.invoke(action)
        tool_message = ToolMessage(
            content = str(response),
            name = action.tool,
            tool_call_id = tool_call["id"]
        )
        return {"messages":[tool_message]}
        
    @abstractmethod
    def get_graph(self) -> CompiledStateGraph:
        pass
    @abstractmethod
    def human_action(self,graph,thread_id=None):
        pass

    def graph_stream(self,graph:CompiledStateGraph,message,thread_id=None):    
        if not thread_id:
            thread_id = self.thread_id
        messages = []
        if message:
            messages.append(HumanMessage(content=message))
            values = {"messages":messages}          
        else:
            values = None
        events = graph.stream( values,
                                config = {"configurable":{"thread_id":thread_id}},
                                stream_mode = "values")
        _print_set = set()
        for event in events:
            self._print_event(event,_print_set)
    
    def _print_event(self, event: dict, _printed: set, max_length=1500):
        current_state = event.get("dialog_state")
        if current_state:
            print("Currently in: ", current_state[-1])
        message = event.get("messages")
        if message:
            if isinstance(message, list):
                message = message[-1]
            if message.id not in _printed:
                msg_repr = message.content
                if len(msg_repr) > max_length:
                    msg_repr = msg_repr[:max_length] + " ... (truncated)"
                if isinstance(message,AIMessage):
                    if message.tool_calls:
                        print(f"{Color.LBLUE}AI:{Color.RESET}",f'使用{Color.GREEN}{message.tool_calls[0]["name"]}{Color.RESET},调用参数:{Color.GREEN}{message.tool_calls[0]["args"]}{Color.RESET}')
                    else:
                        response_metadata = message.response_metadata 
                        print(f"{Color.LBLUE}AI:{Color.RESET}",msg_repr,
                            f'[model:{Color.LYELLOW}{message.response_metadata.get("model_name")}{Color.RESET},token:{Color.LYELLOW}{message.response_metadata.get("token_usage",{}).get("total_tokens")}{Color.RESET}]')
                elif isinstance(message,ToolMessage):
                    print(f"    {Color.BLUE}Tool:{Color.RESET}",msg_repr)
                elif isinstance(message,HumanMessage):
                    print(f"{Color.BLUE}User:{Color.RESET} {msg_repr}")
                _printed.add(message.id)
                
    def graph_get_state_history(self,graph:CompiledStateGraph,thread_id=None):
        if not thread_id:
            thread_id = self.thread_id
        state_history = graph.get_state_history(config = {"configurable":{"thread_id":thread_id}} )
        for state in state_history:
            print("Num Messages: ", len(state.values["messages"]), "Next: ", state.next)
            print("-" * 80)
        return state_history
    
    def graph_get_state(self,graph:CompiledStateGraph,thread_id=None):
        if not thread_id:
            thread_id = self.thread_id
        state = graph.get_state(config =  {"configurable":{"thread_id":thread_id}})
        return state
    
    def graph_update_state(self,graph:CompiledStateGraph,values,thread_id=None,as_node = None):
        if not thread_id:
            thread_id = self.thread_id
        print("as_node:",as_node,"thread_id:",thread_id,"values",values)
        graph.update_state(config = {"configurable":{"thread_id":thread_id}},values=values,as_node=as_node)
    
    def export_graph(self,graph:CompiledStateGraph):
        FileLib.writeFile("graph.png",graph.get_graph().draw_mermaid_png(),mode="wb")  

    


from pydantic import BaseModel,Field
from typing import List,Optional,Literal, Union

from langgraph.graph import MessagesState

class SubTag(BaseModel):
    """
    拆分句子后得到的类型标记。
    diet包括各种饮食，如牛奶、苹果、西红柿炒蛋
    sport包括各种运动，如打篮球、打乒乓、跑步、俯卧撑
    sign包括身体体征，如身高、体重、血压、心率
    buy包括各种购物行为
    """
    sub_text: str = Field(description="相关的子句,注意：需要理解原句子的意思，适当的时候对字句补充相关时间、地点、花费的信息")
    type: Literal["diet","sport","sign","buy","other"] = Field(description="分析句子是关于饮食、运动、体征、购物还是其他")
class Tag(BaseModel):
    '''
    标记句子的意图及类型,并为可以直接给出问题答案的程度打分。要求：
    1、不要遗漏日期、地点等关键条件
    2、如果没有指定日期，且根据上下文也无法判别日期或日期范围，则默认为当天
    举例:
    我吃了一个苹果 -> is_question=fale,action=record,type=diet
    我下午3点要开会 -> is_question=false,action=schedule,type=other
    我解析来一个礼拜要每天跑2公里 -> is_question=false,action=plan,type=sport
    体重70公斤，身高1.74 -> is_question=false,action=recored,type=sign
    怎样学好数学？ -> is_question=true,action=other,type=other
    练习打羽毛球并没有那么容易 -> is_question=false,action=other,type=sport
    下午3点打篮球打到6点半 -> is_question=false,action=record,type=sport
    昨天下午在大润发花了25块3毛六买了3个电风扇 -> is_question=false,action=record,type=buy
    我今天买了什么？ -> is_question=true,action=other,type=buy
    近三天买笔花了多少？ -> is_question=true,action=other,type=buy
    这个礼拜做了哪些运动？-> is_question=true,action=other,type=sport
    这个月平均体重是多少 -> is_question=true,action=other,type=sign 
    今天体重多少公斤？ -> is_question=true,action=other,type=sign
    上周3吃了什么？ -> is_question=true,action=other,type=diet
    除了火龙果，我还吃了什么？-> is_question=true,action=other,type=diet
    '''
    is_question: bool = Field(description="是否为提问语句")
    action: Literal["record","plan","schedule","other"] = Field(description="语句意图是记录(已经完成)、计划(准备完成的目标)还是日程安排(准备完成的安排),如果都不是则为其他")
    subTags: List[SubTag] = Field(description="将语句的原始意思进行拆分。注意：需要理解原句子的意思，适当的时候对字句补充相关时间、地点、花费的信息")
    score: int = Field(description="为可以直接给出问题的答案的可能性打分",min=1,max=5)

class State(MessagesState):
    life_tag: Tag
    human:bool

class Diet(BaseModel):
    '''解析为饮食的相关数据'''
    sdt:Optional[str] = Field(description="吃食品的开始时间,比如上周末、2个小时前、昨天下午2点等")
    edt:Optional[str] = Field(description="吃食品的结束时间,比如上周末、2个小时前、昨天下午2点等")
    duration: Optional[str] = Field(description="吃食品的时长,以分钟单位。如2个小时、3分钟等")
    act:Optional[str] = Field(description="吃食品的动作,例如:吃、喝、口服等")
    place:Optional[str] = Field(description="吃食品的地点")
    name:str = Field(description="食品的名称")
    value:float = Field(description="食品的数量")
    unit: str = Field(description="食品的数量的单位")
    brand: Optional[str] = Field(description="食物的品牌,比如:匹克薯片->brand:匹克,麦当劳汉堡->brand:麦当劳。如果没有指定则为空")
    buy: Optional[float]= Field(description="买食物花的钱，比如：3块5")
    cal: float = Field(description="估算这些数量食物的热量，单位是卡路里。例如:如果一个苹果95卡,那么3个苹果为285卡")
class Diets(BaseModel):
    foods:List[Diet] = Field(description="一组食品")

class Sport(BaseModel):
    '''解析为运动的数据'''
    sdt:Optional[str] = Field(description="运动的开始时间,比如上周末、2个小时前、昨天下午2点等")
    edt:Optional[str] = Field(description="运动的结束时间,比如上周末、2个小时前、昨天下午2点等")
    duration: Optional[str] = Field(description="运动的时间间隔,以分钟单位。比如2个小时、3分钟等")
    act:Optional[str] = Field(description="运动的动作,比如:打、跑、踢、跳等")
    place:Optional[str] = Field(description="运动的地点")
    name:str = Field(description="运动的名称")
    value: Optional[float] = Field(description="运动的数量,比如300次,150米等")
    unit: Optional[str] = Field(description="运动的数量的单位")
    buy: Optional[float]= Field(description="本次运动所花的钱,比如:3块5")
    cal: float = Field(description="估算这些数量或运动时长的运动消耗的热量，单位是卡路里")
class Sports(BaseModel):
    sports:List[Sport] = Field(description="一组运动")

class Sign(BaseModel):
    '''解析为体征测量数据，体征数据包括身高、体重、血压、心率、视力、胸围等'''
    sdt:Optional[str] = Field(description="测量的开始时间，比如上周末、2个小时前、昨天下午2点等")
    edt:Optional[str] = Field(description="测量的结束时间，比如上周末、2个小时前、昨天下午2点等")
    duration: Optional[str] = Field(description="测量的时长,如2个小时、3分钟等")
    act:Optional[str] = Field(description="测量的动作,例如:测量等")
    place:Optional[str] = Field(description="测量的地点")
    name:str = Field(description="体征的名称")
    value: Union[float,str] = Field(description="体征结果的数值，如:30、70.7;或者询问体征的内容，如:最大多少、平均数")
    unit: str = Field(description="体征结果数值的单位，如果没有指定则根据名称默认为常识单位")
    buy: Optional[float]= Field(description="本次测量所花的钱,比如:3块5")

class Signs(BaseModel):
    signs:List[Sign] = Field(description="一组体征")

class Buy(BaseModel):
    '''解析为购买商品的相关数据'''
    sdt:Optional[str] = Field(description="购物的开始时间,比如上周末、2个小时前、昨天下午2点等")
    edt:Optional[str] = Field(description="购物的结束时间,比如上周末、2个小时前、昨天下午2点等")
    duration: Optional[str] = Field(description="购物的时长,以分钟单位。如2个小时、3分钟等")
    act:Optional[str] = Field(description="购物的动作,例如:买、购买等")
    place:Optional[str] = Field(description="购物的地点")
    name:str = Field(description="物品的名称")
    value:float = Field(description="物品的数量")
    unit: str = Field(description="物品的数量的单位")
    brand: Optional[str] = Field(description="物品的品牌,比如:匹克薯片->brand:匹克,麦当劳汉堡->brand:麦当劳。如果没有指定则为空")
    buy: Optional[float]= Field(description="购物花的钱，比如：3块5")
class Buys(BaseModel):
    buys:List[Buy] = Field(description="一组购物")


from  .state import Diets,Sports,Signs,Buys,State
from .node import Node

from langchain_core.messages import AIMessage
class LifeNode(Node):
    def __init__(self,lifeGraph,msg=None):
        super().__init__(lifeGraph,msg)
        self.llm = self.graphLib.get_node_llm()
        self.llm_diet = self.llm.with_structured_output(Diets)
        self.llm_sport = self.llm.with_structured_output(Sports)
        self.llm_sign = self.llm.with_structured_output(Signs)
        self.llm_buy = self.llm.with_structured_output(Buys)
    def __call__(self,state:State):
        message = state["messages"][-1]
        prompt = self.graphLib.langchainLib.get_prompt()
        subTag = state["life_tag"].subTags[0]
        state["life_tag"].subTags = state["life_tag"].subTags[1:]
        tag_type = subTag.type
        match tag_type:
            case "diet":
                res = (prompt | self.llm_diet).invoke({"input":subTag.sub_text})
                if isinstance(res,Diets):
                    self.create_diet_record(res)    
                    return {"messages":[AIMessage(content=str(res))],"life_tag":state["life_tag"]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我无法解析饮食数据")],"life_tag":state["life_tag"]}  
            case "sport":
                res = (prompt | self.llm_sport).invoke({"input":subTag.sub_text})
                if isinstance(res,Sports):
                    self.create_sport_record(res)    
                    return {"messages":[AIMessage(content=str(res))],"life_tag":state["life_tag"]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我无法解析运动数据")],"life_tag":state["life_tag"]}  
            case "sign":
                res = (prompt | self.llm_sign).invoke({"input":subTag.sub_text})
                if isinstance(res,Signs):
                    self.create_sign_record(res)    
                    return {"messages":[AIMessage(content=str(res))],"life_tag":state["life_tag"]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我无法解析体征测量的数据")],"life_tag":state["life_tag"]}  
            case "buy":
                res = (prompt | self.llm_buy).invoke({"input":subTag.sub_text})
                if isinstance(res,Buys):
                    self.create_buy_record(res)    
                    return {"messages":[AIMessage(content=str(res))],"life_tag":state["life_tag"]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我无法解析购物数据")],"life_tag":state["life_tag"]}  
            case _:
                raise Exception("不可能到达这个节点")
        
    def create_diet_record(self,diets:Diets):
        # 处理时间问题
        for diet in diets.foods:
           diet.sdt ,diet.edt, diet.duration = self.parse_time(diet.sdt,diet.edt,diet.duration) 

        diets_list = diets.dict()["foods"]
        user_id = self.graphLib.user_id 
        script = """
            unwind $diets as diet
            match (n:Person{name:$user_id})
            call{            
                with diet with diet
                where diet.name is not null and diet.name <> ""
                merge (f:Food{name:diet.name})
                return f
            }
            call{
                with diet,f
                with diet,f
                where diet.brand is not null and diet.brand <> ""
                merge (b:Brand{name:diet.brand})
                merge (b)-[r:product]->(f) 
            }
            call{
                with diet,n with diet,n
                where diet.place is not null and diet.place <> ""
                merge (p:Place{name:diet.place})
                create (p)<-[r:at{sdt:diet.sdt,edt:diet.edt}]-(n) 
            }
            call{
                with diet,n,f
                create (n)-[r:diet{sdt:diet.sdt,edt:diet.edt,duration:diet.duration,
                    place:diet.place,act:diet.act,name:diet.name,value:diet.value,unit:diet.unit,buy:diet.buy,cal:diet.cal}]->(f)
            }
        """
        self.graphLib.neo4jLib.run(script,diets=diets_list,user_id=user_id)
    
    def create_sport_record(self,sports:Sports):
        # 处理时间问题
        for sport in sports.sports:
            sport.sdt ,sport.edt, sport.duration = self.parse_time(sport.sdt,sport.edt,sport.duration) 

        sports_list = sports.dict()["sports"]
        user_id = self.graphLib.user_id 
        script = """
            unwind $sports as sport
            match (n:Person{name:$user_id})
            call{            
                with sport with sport
                where sport.name is not null and sport.name <> ""
                merge (s:Sport{name:sport.name})
                return s
            }
            call{
                with sport,n with sport,n
                where sport.place is not null and sport.place <> ""
                merge (p:Place{name:sport.place})
                create (p)<-[r:at{sdt:sport.sdt,edt:sport.edt}]-(n) 
            }
            call{
                with sport,n,s
                create (n)-[r:sport{sdt:sport.sdt,edt:sport.edt,duration:sport.duration,
                    place:sport.place,act:sport.act,name:sport.name,value:sport.value,unit:sport.unit,buy:sport.buy,cal:sport.cal}]->(s)
            }
        """
        self.graphLib.neo4jLib.run(script,sports=sports_list,user_id=user_id)
    
    def create_sign_record(self,signs:Signs):
        # 处理时间问题
        for sign in signs.signs:
           sign.sdt ,sign.edt, sign.duration = self.parse_time(sign.sdt,sign.edt,sign.duration) 

        signs_list = signs.dict()["signs"]
        user_id = self.graphLib.user_id 
        script = """
            unwind $signs as sign
            match (n:Person{name:$user_id})
            call{            
                with sign with sign
                where sign.name is not null and sign.name <> ""
                merge (s:Sign{name:sign.name})
                return s
            }
            call{
                with sign,n with sign,n
                where sign.place is not null and sign.place <> ""
                merge (p:Place{name:sign.place})
                create (p)<-[r:at{sdt:sign.sdt,edt:sign.edt}]-(n) 
            }
            call{
                with sign,n,s
                create (n)-[r:sign{sdt:sign.sdt,edt:sign.edt,duration:sign.duration,place:sign.place,act:sign.act,name:sign.name,value:sign.value,unit:sign.unit,buy:sign.buy}]->(s)
            }
        """
        self.graphLib.neo4jLib.run(script,signs=signs_list,user_id=user_id)
    
    def create_buy_record(self,buys:Buys):
        # 处理时间问题
        for buy in buys.buys:
           buy.sdt ,buy.edt, buy.duration = self.parse_time(buy.sdt,buy.edt,buy.duration) 

        buys_list = buys.dict()["buys"]
        user_id = self.graphLib.user_id 
        script = """
            unwind $buys as buy
            match (n:Person{name:$user_id})
            call{            
                with buy with buy
                where buy.name is not null and buy.name <> ""
                merge (pd:Product{name:buy.name})
                return pd
            }
            call{
                with buy,pd
                with buy,pd
                where buy.brand is not null and buy.brand <> ""
                merge (b:Brand{name:buy.brand})
                merge (b)-[r:product]->(pd) 
            }
            call{
                with buy,n with buy,n
                where buy.place is not null and buy.place <> ""
                merge (p:Place{name:buy.place})
                create (p)<-[r:at{sdt:buy.sdt,edt:buy.edt}]-(n) 
            }
            call{
                with buy,n,pd
                create (n)-[r:buy{sdt:buy.sdt,edt:buy.edt,duration:buy.duration,place:buy.place,act:buy.act,name:buy.name,value:buy.value,unit:buy.unit,buy:buy.buy}]->(pd)
            }
        """
        self.graphLib.neo4jLib.run(script,buys=buys_list,user_id=user_id)


from __future__ import annotations
from typing import TYPE_CHECKING, Tuple
if TYPE_CHECKING:
    from ylz_utils.langchain.vectorstores import VectorstoreLib
import logging

try:
    import chromadb
    from langchain_chroma import Chroma
except:
    logging.warning(f"use `pip install langchain-chroma` ,otherwise you can't use provider by ChromaLib")

from langchain_community.docstore.in_memory import InMemoryDocstore
from langchain_core.documents import Document

from typing import List
from uuid import uuid4

class ChromaLib():
    def __init__(self,vectorstoreLib:VectorstoreLib):
        self.vectorstoreLib = vectorstoreLib
        self.config = vectorstoreLib.langchainLib.config
        self.host = self.config.get("VECTORSTORE.CHROMA.HOST")
        self.port = self.config.get("VECTORSTORE.CHROMA.PORT")
        self.db_file = self.config.get("VECTORSTORE.CHROMA.DB_FILE")
        self.server = self.config.get("VECTORSTORE.CHROMA.SERVER")
    def get_store(self,collection_name=None,embedding=None) -> Chroma:
        if not embedding:
            embedding = self.vectorstoreLib.langchainLib.get_embedding()
        if self.server:
            client = chromadb.HttpClient(host=self.host,port=self.port)
            vector_store = Chroma(
                collection_name=collection_name,
                embedding_function=embedding,
                client=client,  # Where to save data locally, remove if not neccesary
            )
        else:
            vector_store = Chroma(
                collection_name=collection_name,
                embedding_function=embedding,
                persist_directory=self.db_file,  # Where to save data locally, remove if not neccesary
            )
        return vector_store
            
    def add_docs(self,vector_store: Chroma,docs,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(docs,batch,vector_store.add_documents)
        return all_ids

    def add_texts(self,vector_store: Chroma,texts,batch:int=1) -> list[str]:
        all_ids = self.vectorstoreLib._split_batch_and_add(texts,batch,vector_store.add_texts)
        return all_ids
    
    def delete(self,vectorstore: Chroma,ids: List[str] | None = None):
        return vectorstore.delete(ids) 
          
    def search(self,query,vectorstore: Chroma,k=10,filter={}):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore: Chroma,k=10,filter={}):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)



from ylz_utils.langchain import LangchainLib
from ylz_utils.data import Spinner, StringLib

from langchain_core.messages import HumanMessage

def input_with_readline(prompt):
    try:
        return input(prompt)
    except UnicodeDecodeError:
        print("输入的内容存在编码问题，请确保使用 UTF-8 编码的字符，并不要使用回退键。")
        return input_with_readline(prompt)
    
def start_chat(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    message = args.message
    model = args.llm_model
    user_id = args.user if args.user else 'default'
    conversation_id = args.conversation if args.conversation else 'default'
    llm = langchainLib.get_llm(key=llm_key,model=model)
    dbname = args.chat_dbname or 'chat.sqlite'
    #### chat 模式
    prompt = langchainLib.get_prompt(use_chat=True)
    if dbname:
        langchainLib.llmLib.set_dbname(dbname)
    chat = langchainLib.get_chat(llm,prompt)
    chain = chat | langchainLib.get_outputParser()
    images=[]
    while True:
        if not message:
            message=input_with_readline(f"{user_id}用户: ")
        else:
            print(f"USER: {message}")
        message = message.strip()
        if message.lower() == '/h':
            print(StringLib.lyellow("/q:"),"退出",end=',')
            print(StringLib.lyellow("/c:"),"清空记录",end=',')
            print(StringLib.lyellow("/u <userid>:"),"设置用户id",end=',')
            print(StringLib.lyellow("/m <image_url>:"),"增加图片",end=',')
            print(StringLib.lyellow("/flux <描述想要生成的图像>:"),"生成图片")
            message = ""
            continue
        if message.lower() in ['/quit','/exit','/stop','/bye','/q']:
            print("Good bye!!")
            break
        if message.lower() == '/c':
            prompt = langchainLib.get_prompt(use_chat=True)
            message = ""
            continue
        if message.lower().startswith('/u'):
            new_user_id = message.split(" ")[1]
            if new_user_id:
                user_id = new_user_id
            message=""
            continue
        if message.lower().startswith('/m'):
            image_url = message.split(" ")[1]
            if image_url:
                images.append({"image":image_url})
            message = ""
            continue
        if message.lower().startswith('/flux'):
            image_message = message.split(" ")[1]
            if image_message:
                spinner = Spinner()
                spinner.start()
                file_name = langchainLib.fluxLib.gen(image_message)
                response = f"已生成图像{file_name},请查看"
                spinner.end()
                print(StringLib.color(response,["blue"]))
            message = ""
            continue
        if images:
            humanMessage = HumanMessage(content = images + [{"text":message}])
            res = llm.stream([humanMessage],{'configurable': {'user_id': user_id,'conversation_id': conversation_id}})
            print("AI: ",end="")
            try:
                for chunk in res:
                    print(chunk.content[0]['text'],end="")
            except:
                print("error!!!!")
                print(type(chunk))
                print(chunk)
            images = []
            message = ""
            print("\n")        
        else:
            res = chain.stream({"input":message}, {'configurable': {'user_id': user_id,'conversation_id': conversation_id}} )
            images = []
            message = ""
            print("AI: ",end="")
            for chunk in res:
                print(chunk,end="")
            print("\n")
    print("\n",langchainLib.get_llm(full=True))


from ylz_utils.data import StringLib
from ylz_utils.langchain import LangchainLib
from langchain_core.messages import HumanMessage,ToolMessage
from ylz_utils.langchain.graph.db_graph import DbGraph
from ylz_utils.langchain.graph.engineer_graph import EngineerGraph
from ylz_utils.langchain.graph.life_graph import LifeGraph
from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph
from ylz_utils.langchain.graph.stand_graph import StandGraph

def input_with_readline(prompt):
    try:
        return input(prompt)
    except UnicodeDecodeError:
        print("输入的内容存在编码问题，请确保使用 UTF-8 编码的字符，并不要使用回退键。")
        return input_with_readline(prompt)
    
def start_graph(langchainLib:LangchainLib,args):
    # # 设置标准输入为 UTF-8 编码
    # sys.stdin = codecs.getreader('utf-8')(sys.stdin.detach())
    graph_key = args.graph or 'stand'
    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message
    chat_dbname = args.chat_dbname
    rag_indexname = args.rag_indexname
    query_dbname = args.query_dbname
    user_id = args.user or 'default'
    conversation_id = args.conversation or 'default'
    thread_id = f"{user_id}-{conversation_id}"
    websearch_key = args.websearch
    graphLib = None
    match graph_key:
        case "stand":
            graphLib = StandGraph(langchainLib)
        case "life":
            graphLib = LifeGraph(langchainLib)
        case "engineer":
            graphLib = EngineerGraph(langchainLib)
        case "db":
            graphLib = DbGraph(langchainLib)
        case "selfrag":
            graphLib = SelfRagGraph(langchainLib)
        case _ :
            return
    if chat_dbname:                                  
        graphLib.set_chat_dbname(chat_dbname)
        print("!!!",f"使用对话数据库{chat_dbname}")
    if rag_indexname:
        retriever = langchainLib.vectorstoreLib.get_store_with_provider_and_indexname(rag_indexname).as_retriever()
        graphLib.set_ragsearch_tool(retriever)
        print("!!!",f"使用知识库{rag_indexname}")
    if query_dbname:
        if query_dbname=="Chinook.db":
            print(StringLib.color(
                "执行wget https://storage.googleapis.com/benchmarks-artifacts/chinook/Chinook.db,确保Chinook.db在当前目录。",
                ["lmagenta","underline"]))
        graphLib.set_query_dbname(query_dbname)
        print("!!!",f"使用查询数据库{query_dbname}")
    if websearch_key:
        graphLib.set_websearch_tool(websearch_key)
        print("!!!",f"使用搜索工具{websearch_key}")
    graphLib.set_nodes_llm_config((llm_key,llm_model))
    graphLib.set_thread(user_id,conversation_id)
    graph = graphLib.get_graph()
    while True:
        if not message:
            #message = input("User Input: ")
            message = input_with_readline("User Input: ")
        else:
            print(f"User:{message}")
        if message.lower() in ["/quit", "/exit", "/stop","/q","/bye"]:
            print("Goodbye!")
            break
        if message=="@@NONE@@":
            message = None
        graphLib.graph_stream(graph,message,thread_id = thread_id)
        if graphLib.human_action(graph,thread_id):
            message = "@@NONE@@"
            continue
        message = ""

    current_state = graphLib.graph_get_state(graph,thread_id)
    print("\n本次对话的所有消息:\n",current_state.values["messages"])

    #langchainLib.graphLib.export_graph(graph)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

import pprint
from langchain_core.prompts import ChatPromptTemplate

class GradeGenerationVDocumentsAndQuestionEdge():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph

    def get_hallucination_grader(self):
        # Prompt
        system = """You are a grader assessing whether an LLM generation is grounded in / supported by a set of retrieved facts. \n 
            Give a binary score 'yes' or 'no'. 'Yes' means that the answer is grounded in / supported by the set of facts."""
        hallucination_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", system),
                ("human", "Set of facts: \n\n {documents} \n\n LLM generation: {generation}"),
            ]
        )

        hallucination_grader = hallucination_prompt | self.self_rag_graph.hallucinations_llm_grader
        #hallucination_grader.invoke({"documents": docs, "generation": generation})
        return hallucination_grader

    def get_answer_grader(self):
        # Prompt
        system = """You are a grader assessing whether an answer addresses / resolves a question \n 
            Give a binary score 'yes' or 'no'. Yes' means that the answer resolves the question."""
        answer_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", system),
                ("human", "User question: \n\n {question} \n\n LLM generation: {generation}"),
            ]
        )

        answer_grader = answer_prompt | self.self_rag_graph.answer_llm_grader
        #answer_grader.invoke({"question": question, "generation": generation})
        return answer_grader
    
    def __call__(self,state):
        """
        Determines whether the generation is grounded in the document and answers question.

        Args:
            state (dict): The current graph state

        Returns:
            str: Decision for next node to call
        """

        print("---CHECK HALLUCINATIONS---")
        question = state["question"]
        documents = state["documents"]
        generation = state["generation"]

        
        score = self.get_hallucination_grader().invoke(
            {"documents": documents, "generation": generation}
        )
        grade = score.binary_score

        # Check hallucination
        if grade == "yes":
            print("---DECISION: GENERATION IS GROUNDED IN DOCUMENTS---")
            # Check question-answering
            print("---GRADE GENERATION vs QUESTION---")
            score = self.get_answer_grader().invoke({"question": question, "generation": generation})
            grade = score.binary_score
            if grade == "yes":
                print("---DECISION: GENERATION ADDRESSES QUESTION---")
                return "useful"
            else:
                print("---DECISION: GENERATION DOES NOT ADDRESS QUESTION---")
                return "not useful"
        else:
            pprint("---DECISION: GENERATION IS NOT GROUNDED IN DOCUMENTS, RE-TRY---")
            return "not supported"


from typing import Optional,Union,Literal
from ylz_utils.config import Config
import logging
import re
from neo4j import GraphDatabase, Result,EagerResult,Record
from neo4j.graph import Node,Relationship

from ylz_utils.data import StringLib

class Neo4jLib():
    def __init__(self,host:str=None,user:str=None ,password:str=None,database:str='neo4j'):
        self.config = Config()
        self.host = host or self.config.get("NEO4J.HOST")
        self.user = user or self.config.get("NEO4J.USER")
        self.password = password or self.config.get("NEO4J.PASSWORD")
        self.database = database or self.config.get("NEO4J.DATABASE")
        self.get_driver()

    def get_driver(self):
        self.driver = GraphDatabase.driver(self.host,auth=(self.user,self.password))
        return self.driver
    
    def close(self):
        if self.driver:
            self.driver.close()

    def run(self,command:str,**kwargs) -> Result:
        if not self.driver:
            self.driver = self.get_driver()
        with self.driver.session(database=self.database) as session:
            try:
                return session.run(command,**kwargs)
            except Exception as e:
                logging.error(e)
                raise e
        
    def query(self,query:str,**kwargs) -> EagerResult:
        if not self.driver:
            self.get_driver()
        try:
            result = self.driver.execute_query(query,database_=self.database,**kwargs) 
            return result   
        except Exception as e:
            logging.error(e)
            raise e
    
    def get_data(self,records:list[Record] | EagerResult):
        if isinstance(records,EagerResult):
            records = records.records
        return [ record.data() for record in records]
    
    def get_rel_schema(self):
        relType = self.get_data(self.query("Call db.schema.relTypeProperties()"))
        return relType
    def get_node_schema(self):
        nodeType = self.get_data(self.query("Call db.schema.nodeTypeProperties()"))
        return nodeType
    '''
        创建主体节点(objects)和客体节点(subjects),主体和客体节点可以重复,以key为唯一标识。
          - 如果objects存在则merge方式创建objects
          - 如果subjects存在则merge方式创建subjects
          - objects与subjects的第一行属性作为节点属性,每一行都应保持同样的结构
          - object_label必须指定
          - 如果subject_label未指定则默认与object_label相同
        创建关系根据relationships,为(object)-[type]->(subject)关系表
          - 如果relationships存在则merge方式创建relation
          - 每一行结构必须包含object,subject,type字段，分别代表主体，客体和关系
          - 第一行属性(除object,subject,type外)作为关系属性，相同type的每一行都应保持同样的结构
          
        例如: 以下数据,key为name
        # 定义Person数据
        objects = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35}
        ]
        subjects = [
            {"name": "Alice", "age": 30},
            {"name": "Bob", "age": 25},
            {"name": "Charlie", "age": 35}
        ]

        # 定义关系数据
        relationships = [
            {"object": "Alice", "subject": "Bob","type":"朋友"},
            {"object": "Alice", "subject": "Charlie","type":"领导"},
            {"object": "Bob", "subject": "Charlie","type":"同事"}
        ]
        create_relationships(objects,subjects,relationships,object_label='Person',subject_label='Person',key='name')
    '''

    def create_node_and_relation(self,data:dict):
        with self.driver.session() as session:
            node_labels = [re.findall(r"(.*)_key",key)[0] for key in data.keys() if re.match(r".*_key",key)] 
            # merge nodes
            for node_label in node_labels:
                nodes = data.get(node_label)
                if nodes:
                    node_properties = "{" + ",".join([f"{key}:node.{key}" for key in nodes[0]]) + "}"
                    create_nodes = f"""
                        UNWIND $nodes AS node
                        MERGE (n:{node_label} {node_properties})
                        """
                    print(StringLib.green(nodes))
                    session.run(create_nodes, nodes=nodes)
            # merge relation
            relations = data.get('relations',[])
            for relation in relations:
                node_from_value = relation.get("from")
                node_from_label = relation.get("from_label")
                node_from_key = relation.get("from_key")
                if not node_from_key:
                    node_from_key = data[f"{node_from_label}_key"]
                node_to_value = relation.get("to")
                node_to_label = relation.get("to_label")
                node_to_key = relation.get("to_key")
                if not node_to_key:
                    node_to_key = data[f"{node_to_label}_key"]
                relation_type = relation["type"]
                relation_prop = [ key for key in relation.keys() if key not in ["from","from_label","from_key","to","to_label","to_key"]]
                relation_properties = "{" + ",".join([f"{key}:rel.{key}" for key in relation_prop]) + "}"
                           
                if node_from_value and node_to_value and node_from_label and node_to_label and relation_type:
                    create_relations = f"""
                        WITH $relation as rel
                        MATCH (a:{node_from_label} {{{node_from_key}: "{node_from_value}"}}), (b:{node_to_label} {{{node_to_key}: "{node_to_value}"}})
                        MERGE (a)-[:{relation_type} {relation_properties}]->(b)
                    """
                    print(StringLib.green(relations))
                    session.run(create_relations,relation=relation)    



from fastapi import FastAPI
from langserve import add_routes
from pydantic import BaseModel,Field
from typing import List,Union
from langchain_core.messages import HumanMessage,AIMessage,SystemMessage
import uvicorn
from fastapi.middleware.cors import CORSMiddleware

from ylz_utils.database.neo4j import Neo4jLib
from ylz_utils.langchain import LangchainLib
from ylz_utils.langchain.graph.life_graph import LifeGraph
from ylz_utils.langchain.graph.test_graph import TestGraph
from langgraph.graph import MessagesState

class InputChat(BaseModel):
    """Input for the chat endpoint."""

    messages: List[Union[HumanMessage, AIMessage, SystemMessage]] = Field(
        ...,
        description="The chat messages representing the current conversation.",
    )
    
def serve(args):
    print("args:",args)
    langchainLib: LangchainLib = LangchainLib()
    langchainLib.add_plugins()    
    path = args.path
    host = args.host
    port = args.port
    llm_key = args.llm_key
    llm_model = args.llm_model

    llm = langchainLib.get_llm(llm_key)
    
    # neo4jLib = Neo4jLib(None,'neo4j','abcd1234')
    # langchainLib.init_neo4j(neo4jLib)
    # lifeGraph = LifeGraph(langchainLib)
    # lifeGraph.set_nodes_llm_config((llm_key,None))
    # lifeGraph.set_thread("youht","default")
    # life_graph = lifeGraph.get_graph()
    
    testGraph = TestGraph(langchainLib)
    testGraph.set_nodes_llm_config({'default':{'llm_key':llm_key,'llm_model':llm_model}})
    test_graph = testGraph.get_graph()

    app = FastAPI(title="Langserve")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
        expose_headers=["*"],
    )
    chain = langchainLib.get_prompt(human_keys={"input":"问题"}) | langchainLib.get_llm(llm_key,llm_model) | langchainLib.get_outputParser()
    add_routes(app,runnable=chain,path=path)

    #add_routes(app,runnable=life_graph,path="/life")
    add_routes(app,runnable=test_graph.with_types(input_type=MessagesState),path="/test_graph",include_callback_events=True)

    uvicorn.run(app, host = host, port = port)



from pydantic import BaseModel,Field
from typing import TypedDict, List

class GradeDocuments(BaseModel):
    """Binary score for relevance check on retrieved documents."""

    binary_score: str = Field(
        description="Documents are relevant to the question, 'yes' or 'no'"
    )

class GradeHallucinations(BaseModel):
    """Binary score for hallucination present in generation answer."""

    binary_score: str = Field(
        description="Answer is grounded in the facts, 'yes' or 'no'"
    )

class GradeAnswer(BaseModel):
    """Binary score to assess answer addresses question."""

    binary_score: str = Field(
        description="Answer addresses the question, 'yes' or 'no'"
    )
    
class GraphState(TypedDict):
    """
    Represents the state of our graph.

    Attributes:
        question: question
        generation: LLM generation
        documents: list of documents
    """

    question: str
    generation: str
    documents: List[str]



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

from langchain import hub
from langchain_core.output_parsers import StrOutputParser

class GenerateNode():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph

    # Post-processing
    def format_docs(docs):
        return "\n\n".join(doc.page_content for doc in docs)
       
    def __call__(self,state):
        """
        Generate answer

        Args:
            state (dict): The current graph state

        Returns:
            state (dict): New key added to state, generation, that contains LLM generation
        """
        print("---GENERATE---")
        question = state["question"]
        documents = state["documents"]
        
        # Prompt
        prompt = hub.pull("rlm/rag-prompt")
        # Chain
        rag_chain = prompt | self.self_rag_graph.llm | StrOutputParser()

        # RAG generation
        generation = rag_chain.invoke({"context": documents, "question": question})
        return {"documents": documents, "question": question, "generation": generation}




from  .state import Tag,State
from .node import Node

from langchain_core.messages import AIMessage

class TagNode(Node):
    def __init__(self,lifeGraph,msg=None):
        super().__init__(lifeGraph,msg)
        self.llm = self.graphLib.get_node_llm()
        self.llm_with_output = self.llm.with_structured_output(Tag)
    def __call__(self,state:State):
        messages= state["messages"]
        message = messages[-1]
        prompt = self.graphLib.langchainLib.get_prompt()
        tag = (prompt | self.llm_with_output).invoke({"input":message.content})
        #tag = self.llm_with_output.invoke(messages)
        if isinstance(tag,Tag):
            return {"life_tag":tag}
        else:
            return {"human":True}  


from .loader import load_github_file
from .llm import _get_llm
from .state import AgentState

prompt = """You are tasked with answering questions about LangGraph functionality and bugs.
Here is a long unit test file for LangGraph. This should contain a lot (but possibly not all) \
relevant information on how to use LangGraph.

<unit_test_file>
{file}
</unit_test_file>

Based on the information above, attempt to answer the user's questions. If you generate a code block, only \
generate a single code block - eg lump all the code together (rather than splitting up). \
You should encode helpful comments as part of that code block to understand what is going on. \
ALWAYS just generate the simplest possible example - don't make assumptions that make it more complicated. \
For "messages", these are a special object that looks like: {{"role": .., "content": ....}}

If users ask for a messages key, use MessagesState which comes with a built in `messages` key. \
You can import MessagesState from `langgraph.graph` and it is a TypedDict, so you can subclass it and add new keys to use as the graph state.

Make sure any generated graphs have at least one edge that leads to the END node - you need to define a stopping criteria!

You generate code using markdown python syntax, eg:

```python
...
```

Remember, only generate one of those code blocks!"""


def draft_answer(state: AgentState, config):
    github_url = "https://github.com/langchain-ai/langgraph/blob/main/libs/langgraph/tests/test_pregel.py"
    file_contents = load_github_file(github_url)
    messages = [
        {"role": "system", "content": prompt.format(file=file_contents)},
                   {"role": "user", "content": state.get('requirements')}
    ] + state['messages']
    model = _get_llm(config,"draft_answer")
    response = model.invoke(messages)
    return {"messages": [response]}



from  .state import Diets,State
from .node import Node

from langchain_core.messages import AIMessage
class AgentNode(Node):
    def __init__(self,lifeGraph,msg=None):
        super().__init__(lifeGraph,msg)
        self.llm = self.graphLib.get_node_llm()
    def __call__(self,state:State):
        messages = state["messages"]
        print("messages--->",messages)
        res = self.llm.invoke(messages)
        return {"messages":[res]}  


from .llm import _get_llm
from .state import AgentState
from typing import TypedDict
from langchain_core.messages import RemoveMessage

gather_prompt = """You are tasked with helping build LangGraph applications. \
LangGraph is a framework for developing LLM applications. \
It represents agents as graphs. These graphs can contain cycles and often contain branching logic.

Your first job is to gather all the user requirements about the topology of the graph. \
You should have a clear sense of all the nodes of the graph/agent, and all the edges. 

You are conversing with a user. Ask as many follow up questions as necessary - but only ask ONE question at a time. \
Only gather information about the topology of the graph, not about the components (prompts, LLMs, vector DBs). \
If you have a good idea of what they are trying to build, call the `Build` tool with a detailed description.

Do not ask unnecessary questions! Do not ask them to confirm your understanding or the structure! The user will be able to \
correct you even after you call the Build tool, so just do enough to get an MVP."""


class Build(TypedDict):
    requirements: str


def gather_requirements(state: AgentState, config):
    messages = [
       {"role": "system", "content": gather_prompt}
   ] + state['messages']
    model = _get_llm(config,"gather_requirements").bind_tools([Build])
    response = model.invoke(messages)
    if len(response.tool_calls) == 0:
        return {"messages": [response]}
    else:
        requirements = response.tool_calls[0]['args']['requirements']
        delete_messages = [RemoveMessage(id=m.id) for m in state['messages']]
        return {"requirements": requirements, "messages": delete_messages}



from typing import Literal
from ylz_utils.data import StringLib
from ylz_utils.file import IOLib

from ylz_utils.langchain.graph import GraphLib

from langgraph.graph.state import CompiledStateGraph
from langgraph.graph import START,END,StateGraph,MessagesState
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage
from langgraph.prebuilt import ToolNode

from langgraph.graph.state import CompiledStateGraph
from langchain_core.messages import ToolMessage
from langgraph.prebuilt import ToolExecutor,ToolInvocation, tools_condition

from .state import State,RequestAssistance
from .chatbot_node import ChatbotNode
from .human_node import HumanNode
from .score_node import ScoreNode

class StandGraph(GraphLib):
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
        
        # tools = [
        #        Tool(name="websearch",
        #             description = "the tool when you can not sure,please search from the internet",
        #             func = websearch
        #        ),
        #        Tool(name="python_repl",
        #             description="when you need to calculation, use python repl tool to execute code ,then return the result to AI.",
        #             func=self.python_repl_tool)
        #      ]
        tools = self.set_tools()
        self.set_tools_executor(tools)

    def chatbot_router(self,state:State)-> Literal["tools","__end__"]:
        messages = state["messages"]
        tool_calls = messages[-1].tool_calls
        if tool_calls:
            return "tools"
        else:
            return END

    def score_router(self,state:State):
        score = state["score"]
        print("score_router.score=",score)
        if score.score <= 1:
            return "chatbotNode"
        else:
            return "humanNode"


    def get_graph(self) -> CompiledStateGraph:
        print("--> create graph:standGraph")
        workflow = StateGraph(State)
        workflow.add_node("scoreNode",ScoreNode(self,"add node:scoreNode"))
        workflow.add_node("chatbotNode",ChatbotNode(self,"add node:chatbotNode"))
        workflow.add_node("toolsNode",self.tool_node)
        workflow.add_node("humanNode", HumanNode(self,"add node:humanNode"))
        
        workflow.add_conditional_edges("chatbotNode", self.chatbot_router,
                                       {"human":"humanNode","tools":"toolsNode","__end__":"__end__"})
        
        workflow.add_edge(START, "scoreNode")
        workflow.add_conditional_edges("scoreNode",self.score_router)
        workflow.add_edge("toolsNode","chatbotNode")
        workflow.add_edge("humanNode","scoreNode")

        graph = workflow.compile(checkpointer=self.memory,
                                 interrupt_before= ["humanNode"])
        
        return graph   

    def human_action(self,graph,thread_id=None) -> bool:
        if not thread_id:
            thread_id = self.thread_id
        snapshot = self.graph_get_state(graph,thread_id)
        if snapshot.next:
            next = snapshot.next[0]
            last_message = snapshot.values["messages"][-1]
            score = snapshot.values['score']
            print(f"    {StringLib.green('Question:')}: {score.request}")
            aiMessage = AIMessage(score.request)
            answer = IOLib.input_with_history(f"    {StringLib.green('Answer:')}")
            humanMessage = HumanMessage(answer)
            snapshot.values["messages"].extend([aiMessage,humanMessage])
            print(snapshot.values["messages"])
            self.graph_update_state(graph,thread_id = thread_id,values=snapshot.values,as_node=next)
            return True
        return False     
 



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.life_graph import LifeGraph

import jionlp
import time
import datetime

class Node():
    def __init__(self,lifeGraph:LifeGraph,msg=None):
        self.graphLib = lifeGraph
        if msg:
            print(msg)
        
    def parse_time(self,sdt,edt,duration):
        parse_sdt=None
        parse_edt=None
        try:
            parse_sdt = jionlp.parse_time(sdt,time_base=time.time())
            print("sdt:",parse_sdt)
        except Exception as e:
            pass
        try:
            parse_edt = jionlp.parse_time(edt,tiem_base=time.time())    
            print("edt:",parse_edt)
        except Exception as e:
            pass
        
        if not parse_sdt and not parse_edt:
            sdt = datetime.datetime.now()
            edt = datetime.datetime.now()
        elif parse_sdt and not parse_edt:
            sdt = datetime.datetime.fromisoformat(parse_sdt['time'][0])
            edt = datetime.datetime.fromisoformat(parse_sdt['time'][1])
        elif parse_edt and not parse_sdt:
            sdt = datetime.datetime.fromisoformat(parse_edt['time'][0])
            edt = datetime.datetime.fromisoformat(parse_edt['time'][1])
        elif parse_edt and parse_sdt:
            sdt = datetime.datetime.fromisoformat(parse_sdt['time'][0])
            edt = datetime.datetime.fromisoformat(parse_edt['time'][1])
        return sdt,edt,duration


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.self_rag_graph import SelfRagGraph

class DecideToGenerateEdge():
    def __init__(self,self_rag_graph:SelfRagGraph):
        self.self_rag_graph = self_rag_graph

    def __call__(self,state):
        """
        Determines whether to generate an answer, or re-generate a question.

        Args:
            state (dict): The current graph state

        Returns:
            str: Binary decision for next node to call
        """

        print("---ASSESS GRADED DOCUMENTS---")
        state["question"]
        filtered_documents = state["documents"]

        if not filtered_documents:
            # All documents have been filtered check_relevance
            # We will re-generate a new query
            print(
                "---DECISION: ALL DOCUMENTS ARE NOT RELEVANT TO QUESTION, TRANSFORM QUERY---"
            )
            return "transform_query"
        else:
            # We have relevant documents, so generate answer
            print("---DECISION: GENERATE---")
            return "generate"




from ylz_utils.langchain.graph import GraphLib

from langgraph.graph import StateGraph, START,END, MessagesState
from langchain_core.messages import AIMessage
from langgraph.graph.state import CompiledStateGraph

from .state import GraphState,GradeDocuments,GradeHallucinations,GradeAnswer

from .node_grade_documents import GradeDocumentsNode
from .node_retrieve import RetrieveNode
from .node_generate import GenerateNode
from .node_transform_query import TransformQueryNode
from .edge_decide_to_generate import DecideToGenerateEdge
from .edge_grade_generation_v_documents_and_question import GradeGenerationVDocumentsAndQuestionEdge


from langchain_community.document_loaders import WebBaseLoader
from langchain_community.vectorstores import Chroma
from langchain_openai import OpenAIEmbeddings

class SelfRagGraph(GraphLib):
    retriever = None
    normal_llm_grader = None
    hallucinations_llm_grader = None
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
    
    def set_retriever(self,retriever=None):
        if retriever:
            self.retriever = retriever
        else:
            dbname = "langgraph1.faiss"
            try:
                vectorstore = self.langchainLib.vectorstoreLib.faissLib.load(dbname)
            except:
                url = "https://langchain-ai.github.io/langgraph/how-tos/"
                docs = self.langchainLib.documentLib.url.load_and_split(url=url,max_depth=1,chunk_size=256,chunk_overlap=0)
                print(f"there is {len(docs)} docs to be add to {dbname}.")
                # Add to vectorDB
                print(docs)
                vectorstore = self.langchainLib.vectorstoreLib.faissLib.get_store()
                _ = self.langchainLib.vectorstoreLib.faissLib.add_docs(vectorstore,docs,batch=10)
                vectorstore.save_local(dbname)
            self.retriever = vectorstore.as_retriever()

    def set_llm(self,llm_key,llm_model):
        self.llm = self.langchainLib.get_llm(llm_key,llm_model)
        self.normal_llm_grader = self.llm.with_structured_output(GradeDocuments)
        self.hallucinations_llm_grader = self.llm.with_structured_output(GradeHallucinations)
        self.answer_llm_grader = self.llm.with_structured_output(GradeAnswer)

    def _check(self):
        if not self.retriever:
            raise Exception("先调用set_retriever(retriever)")
        if not self.llm or not self.normal_llm_grader or \
           not self.hallucinations_llm_grader or \
           not self.answer_llm_grader:
            raise Exception("先调用self.set_llm(llm_key,llm_model)")

    def get_graph(self):
        self._check()
        workflow = StateGraph(GraphState)

        # Define the nodes
        workflow.add_node("retrieve", RetrieveNode(self))  # retrieve
        workflow.add_node("grade_documents", GradeDocumentsNode(self))  # grade documents
        workflow.add_node("generate", GenerateNode(self))  # generatae
        workflow.add_node("transform_query", TransformQueryNode(self))  # transform_query

        # Build graph
        workflow.add_edge(START, "retrieve")
        workflow.add_edge("retrieve", "grade_documents")
        workflow.add_conditional_edges(
            "grade_documents",
            DecideToGenerateEdge(self),
            {
                "transform_query": "transform_query",
                "generate": "generate",
            },
        )
        workflow.add_edge("transform_query", "retrieve")
        workflow.add_conditional_edges(
            "generate",
            GradeGenerationVDocumentsAndQuestionEdge(self),
            {
                "not supported": "generate",
                "useful": END,
                "not useful": "transform_query",
            },
        )

        graph = workflow.compile(self.memory)
        
        return graph


import json
import re
from ylz_utils.file import IOLib
from typing import Literal
from ylz_utils.database.neo4j import Neo4jLib
from ylz_utils.data import StringLib,Spinner
from ylz_utils.langchain import LangchainLib
from langchain_core.messages import AIMessage,HumanMessage
from ylz_utils.langchain.graph.life_graph import LifeGraph
from ylz_utils.langchain.graph.stand_graph import StandGraph

def query_neo4j(neo4jLib:Neo4jLib,vars,query):
    spinner = Spinner()
    try:
        query_variables =re.findall(r"\$(.+?)\s",query)
        kwargs = {}
        for key in query_variables:
            if key in vars:
                kwargs[key] = vars[key]
        spinner.start()
        result = neo4jLib.query(query,**kwargs)
        spinner.end()
        return result
    except Exception as e:
        spinner.end()      
        raise e
    
def neo4j_test(args):
    user = args.user
    password = args.password
    host = args.host
    user_id = args.user_id or 'default'
    conversation_id = args.conversation_id or 'default'
    thread_id = f"{user_id}-{conversation_id}"
    dbname = args.chat_dbname or 'chat.sqlite'
    llm_key = args.llm_key
    llm_model =args.llm_model
    embedding_key = args.embedding_key
    langchainLib = LangchainLib()
    llm = langchainLib.get_llm(llm_key)
    embedding = langchainLib.get_embedding(embedding_key)
    
    prompt = langchainLib.get_prompt(use_chat=True)
    langchainLib.llmLib.set_dbname(dbname)
    chat = langchainLib.get_chat(llm,prompt)
    
    neo4jLib = Neo4jLib(host,user,password)
    langchainLib.init_neo4j(neo4jLib)

    lifeGraph = None
    life_graph = None
    standGraph = None
    stand_graph = None

    print("*"*50,"let's start","*"*50)
    idx = 0
    vars = {}
    history = []
    mode:Literal['neo4j','llm','life','stand','test'] = 'neo4j'
    query=""

    while True:
        idx +=1
        if not query:
            query = IOLib.input_with_history(f'{StringLib.color(f"输入语句{idx}:")}',history)
        #history.append(query)
        if query.strip().lower().startswith('/'):
            command =  query.strip().lower().split(" ")[0]
            if command not in ["/q","/set","/get","/list","/clear","/load","/query","/import","/history","/mode"]:
                print(
"""
usage:
    <CQL> 执行CQL语句
    /set <key>=<value> 设置变量
    /get <key> 查看key变量
    /list 变量列表
    /clear 清除变量
    /load <file name> as <key> 读取文件内容设置为变量key
    /query <key>=<CQL查询语句> 将查询语句的结果设置为变量key
    /import <import.json> 批量创建object/subject节点，以及他们直接的relationship
        创建节点根据nodes以<node_label>_key为唯一标识。
            - 以merge方式创建nodes
            - nodes的第一行属性作为节点属性,每一行都应保持同样的结构
            - <node_label>_key必须指定
        创建关系根据relations,为(from:from_label{...})-[type]->(to:to_label{...})关系表
            - 如果relations存在则merge方式创建relation
            - 每一行结构必须包含from,from_label字段和可选的from_key
            - 第一行属性(除from,to,from_label,to_label,from_key,to_key,type外)作为关系属性
    /mode <neo4j|llm|life|stand|test >  
        - neo4j Neo4j REPL
        - llm <message> 不带graph的对话
        - life 带life_graph的对话 
        - stand 带stand_graph的对话
        - test 测试graph对话
    /q 退出
"""
                )
            elif query.strip().lower() == '/q':
                print("Good bye!")
                break
            elif query.strip().startswith('/set '):
                kv = query.strip().replace("/set","").strip().split('=')
                try:
                    key =kv[0].strip()
                    value = kv[1].strip()
                    vars[key] = json.loads(value)
                except Exception as e:
                    print(e)

            elif query.strip().startswith('/get '):
                key = query.strip().replace("/get","").strip()
                if key in vars:
                    print(f"{key}={vars[key]}")

            elif query.strip().lower().startswith('/list'):
                print(list(vars.keys()))

            elif query.strip().lower().startswith('/clear'):
                vars = {}

            elif query.strip().startswith('/query '):
                try:
                    commands = query.strip().replace('/query ',"").split("=",1)
                    key = commands[0].strip()
                    query = commands[1].strip()            
                    result = query_neo4j(neo4jLib,vars,query)
                    vars[key] = neo4jLib.get_data(result.records)
                except Exception as e:
                    print(StringLib.lred("ERROR ON QUERY:"),e)

            elif query.strip().startswith('/load'):
                try:
                    commands = query.strip().replace('/load',"").split(" as ")
                    filename = commands[0].strip()
                    key = commands[1].strip()
                    with open(filename,"r") as f:
                        text = f.read()
                        vars[key] = json.loads(text)
                except Exception as e:
                    print(StringLib.lred("ERROR ON LOAD:"),e)

            elif query.strip().startswith('/import'):
                try:
                    filename = query.strip().replace('/import',"").strip()
                    with open(filename,"r") as f:
                        importer = json.load(f)
                        neo4jLib.create_node_and_relation(importer)
                except Exception as e:
                    print(StringLib.lred("ERROR ON IMPORT:"),e)

            elif query.strip().startswith('/history'):
                try:
                    command = query.strip().replace('/history',"").strip()
                    match command:
                        case "clear":
                            history = []
                            idx=0
                        case "":
                            for i,item in enumerate(history):
                                print(f"{i+1} - {item}")
                        case _:
                            if  command.isdigit():
                                goto_idx = int(command) 
                            if goto_idx > 0 and goto_idx <= len(history):
                                query = history[goto_idx - 1]
                                continue
                            else: 
                                all = [(i+1,item) for (i,item) in enumerate(history) if re.findall(command,item)]
                                for i,item in all:
                                    print(f"{i} - {item}")
                except Exception as e:
                    print(StringLib.lred("ERROR ON HISTORY:"),e)

            elif query.strip().startswith('/mode'):
                command = query.strip().replace('/mode',"").strip().lower()
                if command not in ['llm','life','test','neo4j']:
                    print("必须指定llm,life,test,neo4j其中的一种!")                
                else:
                    mode = command
                print('当前mode:',mode)
        elif mode == 'llm':
            spinner = Spinner()
            try:
                command = query
                spinner.start()
                res = chat.stream({"input":command}, {'configurable': {'user_id': user_id,'conversation_id': conversation_id}} )
                spinner.end()
                print("AI: ",end="")
                for chunk in res:
                    print(chunk.content,end='')
                print("")
            except Exception as e:
                spinner.end()
                print(StringLib.lred("ERROR ON LLM:"),e)

        elif mode == 'life':
            if not life_graph:
                lifeGraph = LifeGraph(langchainLib)
                lifeGraph.set_nodes_llm_config((llm_key,None))
                lifeGraph.set_thread(user_id,conversation_id)
                life_graph = lifeGraph.get_graph()

            spinner = Spinner()
            try:
                command = query
                spinner.start()
                res = lifeGraph.graph_stream(life_graph,command,thread_id = thread_id)
                spinner.end()
            except Exception as e:
                spinner.end()
                print(StringLib.lred("ERROR ON LLM:"),e)
        elif mode == 'stand' :
            if not stand_graph:
                standGraph = StandGraph(langchainLib)
                standGraph.set_nodes_llm_config((llm_key,None))
                standGraph.set_thread(user_id,conversation_id)
                stand_graph = standGraph.get_graph()

            spinner = Spinner()
            try:
                if query=="@@NONE@@":
                    command = None
                else:
                    command = query
                spinner.start()
                res = standGraph.graph_stream(stand_graph,command,thread_id = thread_id)
                spinner.end()
                if standGraph.human_action(stand_graph,thread_id):
                    query = "@@NONE@@"
                    continue
            except Exception as e:
                spinner.end()
                print(StringLib.lred("ERROR ON Test:"),e)
        else:
            try:
                records,summary,keys = query_neo4j(neo4jLib,vars,query)
                # Records
                for record in records[:5]:
                    print(record.data())
                if records[5:]:
                    n = len(records) - 5
                    if n > 0:
                        print(".........")
                        for record in records[-min(n,5):]:
                            print(record.data())
                # Summary information
                if summary:
                    print("The query `{query}` returned {records_count} records in {time} ms.".format(
                        query=StringLib.color(summary.query), records_count=StringLib.color(len(records)),
                        time=StringLib.color(summary.result_available_after)
                    ))
            except Exception as e:
                print(StringLib.lred("ERROR ON QUERY:"),e)
        
        query=""

    neo4jLib.close()


from pydantic import BaseModel,Field
from typing import TypedDict,Annotated

from langgraph.graph import add_messages

class RequestAssistance(BaseModel):
    """
    Escalate the conversation to an expert. Use this if you are unable to assist directly or if the user requires support beyond your permissions.
    To use this function, relay the user's 'request' so the expert can provide the right guidance.
    """
    request: str =Field(description="当涉及个人状况的时候，使用礼貌的用语向用户提出明确的要询问的问题")
    score:int = Field(description="对询问该问题的必要性给出0-5的评分",min=0,max=5)  

class State(TypedDict):
    messages: Annotated[list,add_messages]
    score: RequestAssistance




from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ylz_utils.langchain.graph.stand_graph import StandGraph

class Node():
    def __init__(self,standGraph:StandGraph,msg=None):
        self.graphLib = standGraph
        if msg:
            print(msg)



from langchain_core.messages import ToolMessage,AIMessage

from .state import State,RequestAssistance
from .node import Node

class ChatbotNode(Node):
    def __init__(self,graphLib,msg=None):
        super().__init__(graphLib,msg)
        self.llm = self.graphLib.get_node_llm()
        self.llm_with_tools = self.llm.bind_tools(self.graphLib.tools)
    def __call__(self,state:State):
        print("begin invoke....")
        print(state["messages"])
        try:
            # if isinstance(state["messages"][-1],ToolMessage):
            #     return {"messages":[AIMessage(state["messages"][-1].content)]}
            response = self.llm_with_tools.invoke(state["messages"])
            #response = self.standGraph.llm.invoke(state["messages"])
            print("end invoke!!!")
            print("\nChatbotNode",response)
            return {"messages":[response]}
        except Exception as e:
           print("ERROR!!!",state)
           raise e



from ylz_utils.langchain.graph import GraphLib

from typing import Literal

from langgraph.graph import StateGraph, END, MessagesState
from langchain_core.messages import AIMessage
from langgraph.graph.state import CompiledStateGraph

from .check import check
from .critique import critique
from .draft import draft_answer
from .gather_requirements import gather_requirements
from .state import AgentState, OutputState, GraphConfig

class EngineerGraph(GraphLib):
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
        # self.set_nodes_llm_config( {
        #     "draft_answer": {"llm_key": "LLM.GROQ","llm_model":None},
        #     "gather_requirements": {"llm_key": "LLM.GROQ","llm_model":None},
        #     "critique": {"llm_key": "LLM.GROQ","llm_model":None},
        # })
    def route_critique(self,state: AgentState) -> Literal["draft_answer", END]:
        if state['accepted']:
            return END
        else:
            return "draft_answer"
    def route_check(self, state: AgentState) -> Literal["critique", "draft_answer"]:
        if isinstance(state['messages'][-1], AIMessage):
            return "critique"
        else:
            return "draft_answer"
    def route_start(self, state: AgentState) -> Literal["draft_answer", "gather_requirements"]:
        if state.get('requirements'):
            return "draft_answer"
        else:
            return "gather_requirements"
    def route_gather(self, state: AgentState) -> Literal["draft_answer", END]:
        if state.get('requirements'):
            return "draft_answer"
        else:
            return END

    def get_graph(self) -> CompiledStateGraph:
        # Define a new graph
        workflow = StateGraph(AgentState, input=MessagesState, output=OutputState, config_schema=GraphConfig)
        workflow.add_node(draft_answer)
        workflow.add_node(gather_requirements)
        workflow.add_node(critique)
        workflow.add_node(check)
        workflow.set_conditional_entry_point(self.route_start)
        workflow.add_conditional_edges("gather_requirements", self.route_gather)
        workflow.add_edge("draft_answer", "check")
        workflow.add_conditional_edges("check", self.route_check)
        workflow.add_conditional_edges("critique", self.route_critique)
        graph = workflow.compile(checkpointer=self.memory)
        return graph 
    






from .state import State,RequestAssistance
from .node import Node
from langchain_core.runnables import RunnablePassthrough

class ScoreNode(Node):
    def __init__(self,standGraph,msg=None):
        super().__init__(standGraph,msg)
        self.llm = self.graphLib.get_node_llm()

    def __call__(self,state:State):
        messages = state["messages"]
        prompt = self.graphLib.langchainLib.get_prompt(
"""
对以下问题评估是否需要询问用户，如果需要给出要询问的内容，如果不需要询问用户则score=1
""",human_keys={},
        use_chat=True)
        #llm = self.langchainLib.get_llm(model = "llama3-groq-70b-8192-tool-use-preview")
        #llm = self.langchainLib.get_llm(model = "llama3-70b-8192")
        chain = {"history":RunnablePassthrough()} | prompt | self.llm.with_structured_output(RequestAssistance)
        response:RequestAssistance = chain.invoke(messages)

        print(response)
        return {"score":response}
        


import datetime
import re
from  .state import State,SubTag
from .node import Node

from langchain_core.messages import AIMessage
class LifeQueryNode(Node):
    def __init__(self,lifeGraph,msg=None):
        super().__init__(lifeGraph,msg)
        self.llm = self.graphLib.get_node_llm()
    def __call__(self,state:State):
        message = state["messages"][-1]
        subTag = state["life_tag"].subTags[0]
        state["life_tag"].subTags = state["life_tag"].subTags[1:]
        tag_type = subTag.type
        match tag_type:
            case 'diet':
                res = self.query_diet(subTag)
                if res:
                    return {"messages":[AIMessage(content=str(res))]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我没有统计出饮食数据")]}  
            case 'sport':
                res = self.query_sport(subTag)
                if res:
                    return {"messages":[AIMessage(content=str(res))]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我没有统计出运动数据")]}  
            case 'sign':
                res = self.query_sign(subTag)
                if res:
                    return {"messages":[AIMessage(content=str(res))]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我没有统计出体征测量的数据")]}  
            case 'buy':
                res = self.query_buy(subTag)
                if res:
                    return {"messages":[AIMessage(content=str(res))]}
                else:
                    return {"messages":[AIMessage(content="抱歉,我没有统计出购物数据")]}  
            case _:
                raise Exception("不应该执行到这里") 
    def query_diet(self,subTag:SubTag):
        prompt_template = \
"""
现在时间是:{now}
我的名字是:{user_id}
根据neo4j的schema一步一步生成查询句子，确保生成neo4j查询语句script与查询的目标相关，且可以正确执行。
要求1、仅输出script，不要做任何解释
    2、仅包含相关的节点和关系
    3、不要使用neo4j不存在的时间函数,使用date(sdt),date(edt)返回日期
schema:{schema}
"""
        prompt = self.graphLib.langchainLib.get_prompt(prompt_template)
        schema_description = \
"""
(:Person).properties: <<name:姓名[String]>>
(:Food).properties: <<name:食物的名称[String]>>
(:Place).properties: <<name:地点名称[String]>>
[s:diet].properties: <<sdt:起始时间,edt:截止时间,place:吃东西的地点[String],cal:食物摄入的热量,value:食物的数量[Double],unit:食物数量的单位[String],buy:吃食物所花费的金额[Double],duration:吃食物所花费的时间[String]>>
[at].properties:<<sdt:起始时间,edt:截止时间>>
(:Person)-[:diet{{sdt,edt,place,value,unit,buy,duration,cal}}]->(:Food) 某人吃喝某些食物
(:Person)-[:at]->(:Place) 某人在某地
"""
        rel_schema = filter(lambda x:x.get("relType") in [":`diet`",":`at`"],self.graphLib.neo4jLib.get_node_schema())
        node_schema = filter(lambda x:x.get("nodeType") in [":`Person`",":`Food`",":`Place`"],self.graphLib.neo4jLib.get_node_schema())
        schema = f"schema description:\n{schema_description}\nrelationship schema:\n{rel_schema}\nnode schema:\n{node_schema}\n"
        res = (prompt | self.llm).invoke({"user_id":self.graphLib.user_id,"now":datetime.datetime.now(),"schema":schema,"input":subTag.sub_text})
        print("query script1---->",res)
        res = re.findall("```cypher(.*)```",res.content.replace("\n"," "))
        if res:
            res = res[0]
            print("query script2---->",res)
            res = self.graphLib.neo4jLib.query(res)
            res = self.graphLib.neo4jLib.get_data(res)
        print("query script3---->",res)
        return res
    def query_sport(self,subTag:SubTag):
        prompt_template = \
"""
现在时间是:{now}
我的名字是:{user_id}
根据neo4j的schema一步一步生成查询句子，确保生成neo4j查询语句script与查询的目标相关，且可以正确执行。
要求1、仅输出script，不要做任何解释
    2、仅包含相关的节点和关系
    3、不要使用neo4j不存在的时间函数,使用date(sdt),date(edt)返回日期
schema:{schema}
"""
        prompt = self.graphLib.langchainLib.get_prompt(prompt_template)
        schema_description = \
"""
(:Person).properties: <<name:姓名[String]>>
(:Sport).properties: <<name:运动的名称[String]>>
(:Place).properties: <<name:地点名称[String]>>
[s:diet].properties: <<sdt:起始时间,edt:截止时间,place:运动的地点[String],cal:运动消耗的热量[Double],value:某项运动的数量[Double],unit:某项运动数量的单位[String],buy:运动所花费的金额[Double],duration:运动所花费的时间[String]>>
[at].properties:<<sdt:起始时间,edt:截止时间>>
(:Person)-[:sport{{sdt,edt,place,value,unit,buy,duration,cal}}]->(:Sport) 某人完成某项运动
(:Person)-[:at]->(:Place) 某人在某地
"""
        rel_schema = filter(lambda x:x.get("relType") in [":`sport`",":`at`"],self.graphLib.neo4jLib.get_node_schema())
        node_schema = filter(lambda x:x.get("nodeType") in [":`Person`",":`Sport`",":`Place`"],self.graphLib.neo4jLib.get_node_schema())
        schema = f"schema description:\n{schema_description}\nrelationship schema:\n{rel_schema}\nnode schema:\n{node_schema}\n"
        res = (prompt | self.llm).invoke({"user_id":self.graphLib.user_id,"now":datetime.datetime.now(),"schema":schema,"input":subTag.sub_text})
        print("query script1---->",res)
        res = re.findall("```cypher(.*)```",res.content.replace("\n"," "))
        if res:
            res = res[0]
            print("query script2---->",res)
            res = self.graphLib.neo4jLib.query(res)
            res = self.graphLib.neo4jLib.get_data(res)
        print("query script3---->",res)
        return res
    def query_sign(self,subTag:SubTag):
        prompt_template = \
"""
现在时间是:{now}
我的名字是:{user_id}
根据neo4j的schema一步一步生成查询句子，确保生成neo4j查询语句script与查询的目标相关，且可以正确执行。
要求1、仅输出script，不要做任何解释
    2、仅包含相关的节点和关系
    3、不要使用neo4j不存在的时间函数,使用date(sdt),date(edt)返回日期
schema:{schema}
"""
        prompt = self.graphLib.langchainLib.get_prompt(prompt_template)
        schema_description = \
"""
(:Person).properties: <<name:姓名[String]>>
(:Sign).properties: <<name:体征的名称[String]>>
(:Place).properties: <<name:地点名称[String]>>
[s:sign].properties: <<sdt:起始时间,edt:截止时间,place:测量地点[String],value:测量的数据[Double],unit:测量的数据的单位[String],buy:测量所花费的金额[Double],duration:测量所花费的时间[String]>>
[at].properties:<<sdt:起始时间,edt:截止时间>>
(:Person)-[:sign{{sdt,edt,place,value,unit,buy,duration}}]->(:Sign) 某人测量某项身体指标
(:Person)-[:at]->(:Place) 某人在某地
"""
        rel_schema = filter(lambda x:x.get("relType") in [":`sign`",":`at`"],self.graphLib.neo4jLib.get_node_schema())
        node_schema = filter(lambda x:x.get("nodeType") in [":`Person`",":`Sign`",":`Place`"],self.graphLib.neo4jLib.get_node_schema())
        schema = f"schema description:\n{schema_description}\nrelationship schema:\n{rel_schema}\nnode schema:\n{node_schema}\n"
        res = (prompt | self.llm).invoke({"user_id":self.graphLib.user_id,"now":datetime.datetime.now(),"schema":schema,"input":subTag.sub_text})
        print("query script1---->",res)
        res = re.findall("```cypher(.*)```",res.content.replace("\n"," "))
        if res:
            res = res[0]
            print("query script2---->",res)
            res = self.graphLib.neo4jLib.query(res)
            res = self.graphLib.neo4jLib.get_data(res)
        print("query script3---->",res)
        return res

    def query_buy(self,subTag:SubTag):
        prompt_template = \
"""
现在时间是:{now}
我的名字是:{user_id}
根据neo4j的schema一步一步生成查询句子，确保生成neo4j查询语句script与查询的目标相关，且可以正确执行。
要求1、仅输出script，不要做任何解释
    2、仅包含相关的节点和关系
    3、不要使用neo4j不存在的时间函数,使用date(sdt),date(edt)返回日期
schema:{schema}
"""
        prompt = self.graphLib.langchainLib.get_prompt(human_prompt=prompt_template)
        schema_description = \
"""
(:Person).properties: <<name:姓名[String]>>
(:Product).properties: <<name:物品的名称[String]>>
(:Place).properties: <<name:地点名称[String]>>
[s:buy].properties: <<sdt:起始时间,edt:截止时间,place:买东西的地点[String],value:买的东西的数量[Double],unit:买东西的数量的单位[String],buy:买东西所花费的金额[Double],duration:买东西所花费的时间[String]>>
[at].properties:<<sdt:起始时间,edt:截止时间>>
(:Person)-[:diet{{sdt,edt,place,value,unit,buy,duration,cal}}]->(:Product) 某人购买某些物品
(:Person)-[:at]->(:Place) 某人在某地
"""
        rel_schema = filter(lambda x:x.get("relType") in [":`buy`",":`at`"],self.graphLib.neo4jLib.get_node_schema())
        node_schema = filter(lambda x:x.get("nodeType") in [":`Person`",":`Product`",":`Place`"],self.graphLib.neo4jLib.get_node_schema())
        schema = f"schema description:\n{schema_description}\nrelationship schema:\n{rel_schema}\nnode schema:\n{node_schema}\n"
        res = (prompt | self.llm).invoke({"user_id":self.graphLib.user_id,"now":datetime.datetime.now(),"schema":schema,"input":subTag.sub_text})
        res = re.findall("```cypher(.*)```",res.content.replace("\n"," "))
        if res:
            res = res[0]
            print("query script2---->",res)
            res = self.graphLib.neo4jLib.query(res)
            res = self.graphLib.neo4jLib.get_data(res)
        print("query script3---->",res)
        return res




from typing import Literal
from ylz_utils.langchain.graph import GraphLib

from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph.message import add_messages
from langgraph.graph import START,END,StateGraph,MessagesState
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage


from .state import State, Tag
from .tag_node import TagNode
from .life_node import LifeNode
from .life_query_node import LifeQueryNode
from .agent_node import AgentNode

class LifeGraph(GraphLib):
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
        self.neo4jLib = self.get_neo4jLib()

    def get_neo4jLib(self):
        neo4jLib =  self.langchainLib.neo4jLib
        if not neo4jLib:
            raise Exception("请先调用langchainLib.init_neo4j(neo4j)")
        self.neo4jLib = neo4jLib
        return neo4jLib
    
    def human_action(self, graph, thread_id):
        return super().human_action(graph, thread_id)
    
    def router(self,state:State)->Literal["life","life_query","agent","__end__"]:
        tag:Tag = state["life_tag"]
        print("tag=======>",tag)
        if tag.is_question:
            for item in tag.subTags:
                match item.type:
                    case "diet":
                        return "life_query"
                    case "sport":
                        return "life_query"
                    case "sign":
                        return "life_query"
                    case "buy":
                        return "life_query"
                    case _:
                        return "agent"
            return "agent"
        match tag.action:
            case "record":
                for item in tag.subTags:
                    match item.type:
                        case "diet":
                            return "life"
                        case "sport":
                            return "life"
                        case "sign":
                            return "life"
                        case "buy":
                            return "life"
                        case _:
                            return "agent"
                return "agent"
            case _:
                return "agent"
    
    def get_graph(self):
        print("---> create graph:lifeGraph")
        workflow = StateGraph(State)
        
        workflow.add_node("tag",TagNode(self,"add node:tagNode"))
        workflow.add_node("life",LifeNode(self,"add node:lifeNode"))
        workflow.add_node("life_query",LifeQueryNode(self,"add node:lifeQueryNode"))
        workflow.add_node("agent",AgentNode(self,"add node:agentNode"))
        
        workflow.add_edge(START,"tag")
        workflow.add_conditional_edges("life",self.router)
        workflow.add_conditional_edges("life_query",self.router)
        workflow.add_edge("agent",END)
        workflow.add_conditional_edges("tag",self.router)
        
        graph = workflow.compile(self.memory)
        return graph 


from enum import Enum
import logging
import textwrap
from typing import List, Dict, Any, Literal, Tuple
import re
from urllib.parse import urlparse, urlunparse
import sys
import time
import itertools
import threading

class Spinner():
    def __init__(self,cursor=['|', '/', '-', '\\']):
        self.cursor = cursor
        self.stop_event = threading.Event()
    def spinning_cursor(self,duration, stop_event):
                spinner = itertools.cycle(self.cursor)
                end_time = time.time() + duration
                while time.time() < end_time:
                    if stop_event.is_set():
                        break
                    sys.stdout.write(next(spinner))
                    sys.stdout.flush()
                    sys.stdout.write('\b')  # 删除上一个字符
                    time.sleep(0.1)  # 控制动画的速度   
    def start(self):
        self.spinner_thread = threading.Thread(target=self.spinning_cursor, args=(float('inf'), self.stop_event))
        self.spinner_thread.start()
    
    def end(self):
        self.stop_event.set()
        self.spinner_thread.join()

class Color: 
    BLACK = '\033[30m'
    WHITE = '\033[37m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    LRED = '\033[91m'
    LGREEN = '\033[92m'
    LYELLOW = '\033[93m'
    LBLUE = '\033[94m'
    LMAGENTA = '\033[95m'
    LCYAN = '\033[96m'
    RESET = '\033[0m'
    BOLD = '\033[1m'
    DARK = '\033[2m'
    ITALIC = '\033[3m'
    UNDERLINE = '\033[4m'
    DELETE = '\033[9m'
ColorMap = {
    "black":Color.BLACK,"white":Color.WHITE,
    "red":Color.RED,"lred":Color.LRED,
    "green":Color.GREEN,"lgreen":Color.LGREEN,
    "yellow":Color.YELLOW,"lyellow":Color.LYELLOW,
    "blue":Color.BLUE,"lblue":Color.LBLUE,
    "magenta":Color.MAGENTA,"lmagenta":Color.LMAGENTA,
    "cyan":Color.CYAN,"lcyan":Color.LCYAN,
    "reset":Color.RESET,"bold":Color.BOLD,"dark":Color.DARK,"italic":Color.ITALIC,
    "underline":Color.UNDERLINE,"delete":Color.DELETE
}
class MathLib:
    pass

class StringLib:
    @classmethod
    def black(cls,text):
        return f"{Color.BLACK}{text}{Color.RESET}"
    @classmethod
    def white(cls,text):
        return f"{Color.WHITE}{text}{Color.RESET}"
    @classmethod
    def red(cls,text):
        return f"{Color.RED}{text}{Color.RESET}"
    @classmethod
    def green(cls,text):
        return f"{Color.GREEN}{text}{Color.RESET}"
    @classmethod
    def yellow(cls,text):
        return f"{Color.YELLOW}{text}{Color.RESET}"
    @classmethod
    def blue(cls,text):
        return f"{Color.BLUE}{text}{Color.RESET}"
    @classmethod
    def magenta(cls,text):
        return f"{Color.MAGENTA}{text}{Color.RESET}"
    @classmethod
    def cyan(cls,text):
        return f"{Color.CYAN}{text}{Color.RESET}"
    @classmethod
    def lred(cls,text):
        return f"{Color.LRED}{text}{Color.RESET}"
    @classmethod
    def lgreen(cls,text):
        return f"{Color.LGREEN}{text}{Color.RESET}"
    @classmethod
    def lyellow(cls,text):
        return f"{Color.LYELLOW}{text}{Color.RESET}"
    @classmethod
    def lblue(cls,text):
        return f"{Color.LBLUE}{text}{Color.RESET}"
    @classmethod
    def lmagenta(cls,text):
        return f"{Color.LMAGENTA}{text}{Color.RESET}"
    @classmethod
    def lcyan(cls,text):
        return f"{Color.LCYAN}{text}{Color.RESET}"
    @classmethod
    def bold(cls,text):
        return f"{Color.BOLD}{text}{Color.RESET}"
    @classmethod
    def dark(cls,text):
        return f"{Color.DARK}{text}{Color.RESET}"
    @classmethod
    def italic(cls,text):
        return f"{Color.ITALIC}{text}{Color.RESET}"
    @classmethod
    def underline(cls,text):
        return f"{Color.UNDERLINE}{text}{Color.RESET}"
    @classmethod
    def delete(cls,text):
        return f"{Color.DELETE}{text}{Color.RESET}"
    @classmethod
    def color(cls,text,style:\
              List[Literal["black","white","red","lred","green","lgreen","yellow","lyellow",\
                           "blue","lblue","magenta","lmagenta","cyan","lcyan", \
                           "bold","dark","italic","underline","delete"]] = ["yellow"]):
        if not style:
            return text
        for item in style:
            text = f"{ColorMap[item]}{text}{Color.RESET}"
        return text
    @classmethod
    def logging_in_box(cls,text, char="=", console_width:int=80, print_func = logging.info):
        """
        将传入的字符串用“=”字符串框起来在console打印出来，支持多行文本，= 对齐，
        并考虑了字符串在控制台中的实际显示宽度。

        Args:
            text: 要打印的字符串，可以包含多行。
            console_width: 控制台的宽度，默认为 80 个字符。
        """

        lines = text.splitlines()

        # 计算边框长度，考虑控制台宽度
        max_line_length = max(len(line) for line in lines)
        border_length = min(max_line_length + 4, console_width)

        # 打印上边框
        print_func(char * border_length)

        for line in lines:
            # 截断过长的行
            wrapped_lines = textwrap.wrap(line, width=console_width - 4)

            for wrapped_line in wrapped_lines:
                # 计算需要填充的空格数
                padding_space = border_length - len(wrapped_line) - 4
                padding = " " * padding_space

                # 打印带边框的文本行
                print_func(f"{char} {wrapped_line}{padding} {char}")

        # 打印下边框
        print_func(char * border_length)
    
class UrlLib:
    @classmethod
    def urlify(cls, address) -> tuple[str,str]:
        address = address.strip()
        # 解析URL
        parsed_url = urlparse(address)

        # 默认协议为 http
        if not parsed_url.scheme:
            address = 'http://' + address
            parsed_url = urlparse(address)

        # 判断地址类型
        url_type = None
        if parsed_url.scheme == 'file':
            url_type = 'file'
        elif parsed_url.scheme in ['http', 'https']:
            url_type = 'web'
        else:
            url_type = 'unknown'

        return address, url_type
    
    @classmethod
    def strip_protocol(cls, url):
        parsed_url = urlparse(url)
        # 构建去掉协议的URL
        stripped_url = parsed_url.netloc + parsed_url.path
        if parsed_url.query:
            stripped_url += '?' + parsed_url.query
        if parsed_url.fragment:
            stripped_url += '#' + parsed_url.fragment
        return stripped_url


class JsonLib:
    @classmethod
    def find_key_value_path(cls, json_data, target_key_path):
        # target_key_path 可以是以下类型
        # [] 表示数组占位
        # [2] 表示仅第3个数组占位
        # * 表示任何非数组的key占位
        # 其他字符串，表示key
        # example:  name.*.address , public.[].note.[1].local
        keys = target_key_path.split(".")
        keys.reverse()
        results = []
        def _search(data,current_path=[]):
            if isinstance(data,list):
                for index,item in enumerate(data):
                    current_path.append(index)
                    _search(item,current_path)
                    current_path.pop()
            elif isinstance(data,dict):
                for dict_key in data:
                    current_path.append(dict_key)
                    _search(data[dict_key],current_path)
                    current_path.pop()
            else:
                reversed_path = current_path.copy()
                if len(keys) > len(reversed_path):
                    return
                reversed_path.reverse()
                #print(data,keys,reversed_path)
                zip_keys = zip(keys,reversed_path)
                use = True
                for item in zip_keys:
                    key = item[0]
                    path = item[1]
                    if isinstance(path,int):
                        if key=="[]":
                            continue
                        elif re.match(r"\[(\d+)\]",key):
                            num = int(re.findall(r"\[(\d*)\]",key)[0])
                            if num != path:
                                use = False
                                break
                            else:
                                continue
                    if key=="*":
                        continue
                    if key != path:
                        use = False
                        break
                if use:
                    results.append({"value":data,"path":current_path.copy()})
        _search(json_data)
        return results                        
    @classmethod
    def find_key_value_path_old(cls, json_data, target_key_path,no_dict = True):
        """
        在 JSON 数据中查找特定路径下的 key 对应的值，并返回带有 JSON path 的结果。

        Args:
            json_data: 要搜索的 JSON 数据 (字典或列表)。
            target_key_path: 要查找的 key 路径，使用 "." 分隔层级，
                             例如 "address.city" 或 "phoneNumbers.[].city" 或 "bus.[].b.[].b1"。
            no_dict: 如果为True则只会查找最终的值，会忽略字典。如果为False找到匹配的字典不会再深入到字典里检索。默认为True
        Returns:
            一个列表，包含所有找到的 key 的值和路径信息。
            每个元素是一个字典，包含以下键值对：
                - "value": 找到的值
                - "path": 值在 JSON 数据中的路径，例如 ["phoneNumbers", 0, "city"]

        例如：
            [
                {"value": "Anytown", "path": ["address", "city"]},
                {"value": "Hometown", "path": ["phoneNumbers", 0, "city"]},
                {"value": "Workcity", "path": ["phoneNumbers", 1, "city"]}
            ]
        """

        keys = target_key_path.split(".")
        old_keys = "/".join(map(lambda key:'@[]@' if re.match(r'\[.*\]',key) else f"@{key}@",keys))
        results: List[Dict[str, Any]] = []

        def _process_keys(data, keys, current_path: List):
            if not keys:
                new_path = current_path.copy()
                new_keys = "/".join(map(lambda key:'@[]@' if re.match(r'^\d+$',str(key)) else f"@{str(key)}@",new_path))
                if (not isinstance(data, list)) and (not isinstance(data, dict)):
                    if new_keys.endswith(old_keys):
                        #results.append({"value":data,"path": new_path,"o1":old_keys,"n1":new_keys})
                        results.append({"value":data,"path": new_path})
                    return
                else:
                    if no_dict:
                        keys = target_key_path.split(".")
                    else:
                        if new_keys.endswith(old_keys):
                            #results.append({"value":data,"path": new_path,"o2":old_keys,"n2":new_keys})
                            results.append({"value":data,"path": new_path})
                        return
            key = keys[0]
            if key==[]:
                if isinstance(data, list):
                    for i, item in enumerate(data):
                        current_path.append(i)
                        _process_keys(item, keys[1:], current_path)
                        current_path.pop()
            elif re.match(r"\[(\d+)\]",key):
                num = int(re.findall(r"\[(\d+)\]",key)[0])
                if isinstance(data, list):
                    if num < len(data):
                        current_path.append(num)
                        _process_keys(data[num], keys[1:], current_path)
                        current_path.pop()
            else:
                if isinstance(data, dict):
                    if key in data:
                        current_path.append(key)
                        _process_keys(data[key], keys[1:], current_path)
                        current_path.pop()
                    else:
                        for data_key in data.keys():
                            current_path.append(data_key)
                            _process_keys(data[data_key], keys, current_path)
                            current_path.pop()
                elif isinstance(data, list):
                    for i, item in enumerate(data):
                        current_path.append(i)
                        _process_keys(item, keys, current_path)
                        current_path.pop()
                else:
                    new_path = current_path.copy()
                    new_keys = "/".join(map(lambda key:'@[]@' if re.match(r'^\d+$',str(key)) else f'@{str(key)}@',new_path))
                    if new_keys.endswith(old_keys):
                        #results.append({"value":data,"path": new_path,"o3":old_keys,"n3":new_keys})
                        results.append({"value":data,"path": new_path})
        _process_keys(json_data, keys, [])
        return results
    @classmethod
    def update_json_by_path(cls, json_data:dict, updates=[], func=None)->dict:
        # updates = [
        #   {"value":"abc","path":["a","b",1,"c"]}
        #   {"value":"xyz","path":["x","0",y,"z"]}
        #   ...
        # ]
        # 将用func("abc") 更新a.b[1].c
        #    func("xyz") 更新x[0].y.z
        def _update_json_recursive(data, path, value):
            """递归地更新 JSON 数据。"""
            if not path:  # 路径为空，直接返回
                return

            key = path[0]

            if isinstance(data, list) and isinstance(key, int):
                if key >= len(data):
                    data.extend([None] * (key - len(data) + 1))
                if len(path) == 1:
                    data[key] = value
                else:
                    if data[key] is None:
                        data[key] = {}
                    _update_json_recursive(data[key], path[1:], value)
            elif isinstance(data, dict):
                if len(path) == 1:
                    data[key] = value
                else:
                    if key not in data:
                        data[key] = {}
                    _update_json_recursive(data[key], path[1:], value)
            else:
                raise ValueError(f"Unsupported type for key '{key}' at path: {path}")    
        for update in updates:
            value = update['value']
            if func:
                value = func(value)
            path = update['path']
            _update_json_recursive(json_data, path, value)
        return json_data


if __name__ == "__main__":
    json_data = {
        "name": "John",
        "age": 30,
        "address": {
            "street": "123 Main St",
            "city": "Anytown",
            "zip": "12345"
        },
        "p":{
          "a":{"p":5},"b":{"p":6},"c":{"p":7}
        },
        "phoneNumbers": [
            {"type": "home", "number": "555-555-1212", "city": "Hometown"},
            {"type": "work", "number": "555-555-1213", "city": "Workcity"}
        ],
        "bus": [
            {"p":888},
            {"a": [1,3,5], "b": {"b1": [1,{"x":1}],"p":999}},
            {"a": 2, "b": [{"bus": [3,{"y":2}]}]}
        ]
    }
    
    # 测试用例
    results = JsonLib.find_key_value_path(json_data, "city")
    print(f"\n\n结果1 =====> {results}")  # 应该输出：['Anytown', ['Hometown', 'Workcity']]

    
    results = JsonLib.find_key_value_path(json_data, "*.*.p")
    # results = JsonLib.find_key_value_path(json_data, "a",no_dict=False)
    print(f"\n\n结果2 ===>: {results}")  

    # results = JsonLib.update_json_by_path(json_data,results,lambda x:x*2)
    # print("*"*20,"\n",results)

    


from langchain_core.messages import ToolMessage

from .state import State,RequestAssistance
from .node import Node

class HumanNode(Node):
    def __call__(self,state: State):
        score = state["score"]
        score.score=0
        return {"score":score}


from ylz_utils.langchain.graph import GraphLib
from langgraph.graph import StateGraph,MessagesState,START,END
from langchain_core.messages import AIMessage
class TestGraph(GraphLib):
    def __init__(self,langchainLib):
        super().__init__(langchainLib)
    
    def get_graph(self):
        workflow = StateGraph(MessagesState)
        workflow.add_node("nodeA",NodeA("hello",self))
        workflow.add_node("nodeB",NodeB("3+3=?",self))
        workflow.add_edge(START,"nodeA")
        workflow.add_edge("nodeA","nodeB")
        workflow.add_edge("nodeB",END)        
        graph = workflow.compile(self.memory)
        return graph

    def human_action(self, graph, thread_id=None):
        return super().human_action(graph, thread_id)
    
class NodeA():
    def __init__(self,msg,graphLib:GraphLib=None):
        self.llm=None
        if graphLib:
            self.llm = graphLib.get_node_llm()
        self.msg = msg
    def __call__(self,state):
        if self.llm:
            res = self.llm.invoke(self.msg)
            return {"messages":[res]}
        return {"messages":[AIMessage(content=self.msg)]}

class NodeB():
    def __init__(self,msg,graphLib:GraphLib=None):
        self.llm=None
        if graphLib:
            self.llm = graphLib.get_node_llm()
        self.msg = msg
        self.tools = [self.add]
    def add(self,a:int , b:int)->int:
        '''将两个整数相加'''
        return a+b
    def __call__(self,state):
        if self.llm:
            llm_bind = self.llm.bind_tools(self.tools)
            res = llm_bind.invoke(self.msg)
            print("???",res)
            return {"messages":[res]}
        return {"messages":[AIMessage(content=self.msg)]}
    